@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/api.rs: JSON serializers for syd(2) API
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -499,9 +499,18 @@ pub static API_SPEC: LazyLock<Api> = LazyLock::new(|| Api {
             ..Default::default()
         },
         Method {
-            name: "allow_unsafe_nopie".into(),
+            name: "allow_unsafe_sticky".into(),
+            desc: "Disable immutable sticky bit enforcement for directories".into(),
+            path: Some("trace/allow_unsafe_sticky".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_exec_nopie".into(),
             desc: "Allow unsafe execution of non-PIE binaries".into(),
-            path: Some("trace/allow_unsafe_nopie".into()),
+            path: Some("trace/allow_unsafe_exec_nopie".into()),
             argv: Some(vec![Arg::Boolean]),
             argc: Some(vec![1]),
             op: Some(Operator::Set),
@@ -527,8 +536,8 @@ pub static API_SPEC: LazyLock<Api> = LazyLock::new(|| Api {
         },
         Method {
             name: "allow_unsafe_xattr".into(),
-            desc: "Allow unsafe access to sensitive extensive attributes".into(),
-            path: Some("trace/allow_unsafe_open_xattr".into()),
+            desc: "Allow unsafe access to sensitive extended attributes".into(),
+            path: Some("trace/allow_unsafe_xattr".into()),
             argv: Some(vec![Arg::Boolean]),
             argc: Some(vec![1]),
             op: Some(Operator::Set),
@@ -561,6 +570,96 @@ pub static API_SPEC: LazyLock<Api> = LazyLock::new(|| Api {
             op: Some(Operator::Set),
             ..Default::default()
         },
+        Method {
+            name: "allow_unsafe_sendfd_dir".into(),
+            desc: "Allow sending directory file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_dir".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_bdev".into(),
+            desc: "Allow sending block device file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_bdev".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_cdev".into(),
+            desc: "Allow sending character device file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_cdev".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_symlink".into(),
+            desc: "Allow sending symbolic link file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_symlink".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_magiclink".into(),
+            desc: "Allow sending magic symbolic link file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_magiclink".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_memfd".into(),
+            desc: "Allow sending memory file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_memfd".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_secretmem".into(),
+            desc: "Allow sending secret memory file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_secretmem".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_socket".into(),
+            desc: "Allow sending socket file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_socket".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_fifo".into(),
+            desc: "Allow sending FIFO file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_fifo".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_sendfd_misc".into(),
+            desc: "Allow sending unknown/misc file descriptors with SCM_RIGHTS".into(),
+            path: Some("trace/allow_unsafe_sendfd_misc".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
         Method {
             name: "deny_dotdot".into(),
             desc: "Deny .. components in path resolution".into(),
@@ -571,36 +670,36 @@ pub static API_SPEC: LazyLock<Api> = LazyLock::new(|| Api {
             ..Default::default()
         },
         Method {
-            name: "deny_elf32".into(),
+            name: "deny_exec_elf32".into(),
             desc: "Deny execution of 32-bit ELF binaries".into(),
-            path: Some("trace/deny_elf32".into()),
+            path: Some("trace/deny_exec_elf32".into()),
             argv: Some(vec![Arg::Boolean]),
             argc: Some(vec![1]),
             op: Some(Operator::Set),
             ..Default::default()
         },
         Method {
-            name: "deny_elf_dynamic".into(),
+            name: "deny_exec_elf_dynamic".into(),
             desc: "Deny execution of dynamically linked ELF binaries".into(),
-            path: Some("trace/deny_elf_dynamic".into()),
+            path: Some("trace/deny_exec_elf_dynamic".into()),
             argv: Some(vec![Arg::Boolean]),
             argc: Some(vec![1]),
             op: Some(Operator::Set),
             ..Default::default()
         },
         Method {
-            name: "deny_elf_static".into(),
+            name: "deny_exec_elf_static".into(),
             desc: "Deny execution of statically linked ELF binaries".into(),
-            path: Some("trace/deny_elf_static".into()),
+            path: Some("trace/deny_exec_elf_static".into()),
             argv: Some(vec![Arg::Boolean]),
             argc: Some(vec![1]),
             op: Some(Operator::Set),
             ..Default::default()
         },
         Method {
-            name: "deny_script".into(),
+            name: "deny_exec_script".into(),
             desc: "Deny execution of scripts".into(),
-            path: Some("trace/deny_script".into()),
+            path: Some("trace/deny_exec_script".into()),
             argv: Some(vec![Arg::Boolean]),
             argc: Some(vec![1]),
             op: Some(Operator::Set),
@@ -669,6 +768,132 @@ pub static API_SPEC: LazyLock<Api> = LazyLock::new(|| Api {
             op: Some(Operator::Set),
             ..Default::default()
         },
+        Method {
+            name: "allow_unsafe_exec_ldso".into(),
+            desc: "Allow ld.so exec indirection".into(),
+            path: Some("trace/allow_unsafe_exec_ldso".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_exec_stack".into(),
+            desc: "Allow execution of binaries with executable stack".into(),
+            path: Some("trace/allow_unsafe_exec_stack".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_create".into(),
+            desc: "Allow unsafe file creation".into(),
+            path: Some("trace/allow_unsafe_create".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_machine_id".into(),
+            desc: "Allow real machine-id(5) exposure".into(),
+            path: Some("trace/allow_unsafe_machine_id".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_symlinks".into(),
+            desc: "Allow following symlinks in untrusted directories".into(),
+            path: Some("trace/allow_unsafe_symlinks".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_stat_bdev".into(),
+            desc: "Allow stat times on block devices".into(),
+            path: Some("trace/allow_unsafe_stat_bdev".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_stat_cdev".into(),
+            desc: "Allow stat times on character devices".into(),
+            path: Some("trace/allow_unsafe_stat_cdev".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_notify_bdev".into(),
+            desc: "Allow fanotify/inotify for block devices".into(),
+            path: Some("trace/allow_unsafe_notify_bdev".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_notify_cdev".into(),
+            desc: "Allow fanotify/inotify for character devices".into(),
+            path: Some("trace/allow_unsafe_notify_cdev".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "allow_unsafe_any_addr".into(),
+            desc: "Allow wildcard any address connections".into(),
+            path: Some("trace/allow_unsafe_any_addr".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "deny_vdso".into(),
+            desc: "Deny access to vDSO".into(),
+            path: Some("trace/deny_vdso".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "force_local_net".into(),
+            desc: "Force local networking".into(),
+            path: Some("trace/force_local_net".into()),
+            argv: Some(vec![Arg::Boolean]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "log_level".into(),
+            desc: "Set the log level".into(),
+            path: Some("log/level".into()),
+            argv: Some(vec![Arg::Level]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
+        Method {
+            name: "log_verbose".into(),
+            desc: "Set verbose logging level".into(),
+            path: Some("log/verbose".into()),
+            argv: Some(vec![Arg::U8]),
+            argc: Some(vec![1]),
+            op: Some(Operator::Set),
+            ..Default::default()
+        },
     ],
     types: vec![
         Type {
@@ -748,7 +973,30 @@ pub static API_SPEC: LazyLock<Api> = LazyLock::new(|| Api {
         Type {
             name: "$state".into(),
             desc: "Sandbox lock state".into(),
-            enums: Some(vec!["on".into(), "off".into(), "exec".into(), "ipc".into()]),
+            enums: Some(vec![
+                "off".into(),
+                "exec".into(),
+                "ipc".into(),
+                "drop".into(),
+                "read".into(),
+                "on".into(),
+            ]),
+            impls: Some(vec!["$str".into()]),
+            ..Default::default()
+        },
+        Type {
+            name: "$level".into(),
+            desc: "Log level".into(),
+            enums: Some(vec![
+                "emerg".into(),
+                "alert".into(),
+                "crit".into(),
+                "error".into(),
+                "warn".into(),
+                "notice".into(),
+                "info".into(),
+                "debug".into(),
+            ]),
             impls: Some(vec!["$str".into()]),
             ..Default::default()
         },
@@ -1097,6 +1345,7 @@ enum Arg {
     Capability, // sandbox capability
     Rule,       // sandbox rule
     State,      // lock state
+    Level,      // log level
     FileDes,    // file descriptor
     Profile,    // builtin profile
 }
@@ -1127,6 +1376,7 @@ impl Serialize for Arg {
             Self::Capability => serializer.serialize_str("$caps"),
             Self::Rule => serializer.serialize_str("$rule"),
             Self::State => serializer.serialize_str("$state"),
+            Self::Level => serializer.serialize_str("$level"),
             Self::FileDes => serializer.serialize_str("$filedes"),
             Self::Profile => serializer.serialize_str("$profile"),
         }
