@@ -9,6 +9,8 @@
 use libseccomp::ScmpNotifResp;
 use nix::errno::Errno;
 
+#[cfg(feature = "kcov")]
+use crate::kcov::abi::kcov_ioctl;
 use crate::{
     ioctl::Ioctl,
     log::get_ioctl_log,
@@ -25,6 +27,19 @@ pub(crate) fn sys_ioctl(request: UNotifyEventRequest) -> ScmpNotifResp {
         let req = request.scmpreq;
         let sandbox = request.get_sandbox();
 
+        #[cfg(feature = "kcov")]
+        {
+            // Check ioctl(2) is for KCOV.
+            let arg = crate::req::SysArg {
+                dirfd: Some(0),
+                ..Default::default()
+            };
+            let (path, _, _) = request.read_path(&sandbox, arg)?;
+            if path.is_memory_fd() && path.abs().is_kcov_mfd() {
+                return kcov_ioctl(&request);
+            }
+        }
+
         // Check if the ioctl(2) request is allowed or denied.
         let arg = req.data.args[1];
         let cap = Capability::CAP_IOCTL;
