@@ -2,15 +2,26 @@
 // Syd: rock-solid application kernel
 // src/kernel/mod.rs: Secure computing hooks
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
 macro_rules! syscall_handler {
     ($request:ident, $body:expr) => {{
         let request_id = $request.scmpreq.id;
+        let _request_tid = $request.scmpreq.pid();
 
-        match $body($request) {
+        #[cfg(feature = "kcov")]
+        {
+            crate::kcov::abi::kcov_attach(_request_tid);
+            crate::kcov::abi::kcov_set_syscall(
+                $request.scmpreq.data.syscall.as_raw_syscall().into(),
+            );
+            let _ = crate::kcov::abi::kcov_enter_for(_request_tid);
+            crate::kcov_edge!();
+        }
+
+        let result = match $body($request) {
             Ok(result) => result,
             // SAFETY: Harden against UnknownErrno so as not to
             // confuse the Linux API from returning no-op.
@@ -23,7 +34,15 @@ macro_rules! syscall_handler {
                 let errno = (errno as i32).checked_neg().unwrap_or(-libc::ENOSYS);
                 ScmpNotifResp::new(request_id, 0, errno, 0)
             }
+        };
+
+        #[cfg(feature = "kcov")]
+        {
+            crate::kcov_edge!();
+            let _ = crate::kcov::abi::kcov_exit_for(_request_tid);
         }
+
+        result
     }};
 }
 
@@ -135,12 +154,13 @@ pub(crate) mod xattr;
 /// ptrace(2) hooks
 pub(crate) mod ptrace;
 
-use std::borrow::Cow;
+/// ptrace(2) syscall handler
+pub(crate) mod sys_ptrace;
 
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
-    fcntl::{AtFlags, OFlag},
+    fcntl::AtFlags,
     sys::{
         signal::{kill, Signal},
         stat::{Mode, SFlag},
@@ -149,14 +169,15 @@ use nix::{
 };
 
 use crate::{
-    fs::{fd_status_flags, to_valid_fd},
-    ioctl::{ioctl_names_get, Ioctl},
+    fd::to_valid_fd,
     log::log_is_main,
+    log_enabled,
     lookup::{CanonicalPath, FileInfo, FileType},
     notice,
-    path::{XPath, XPathBuf},
+    path::XPath,
     req::{PathArgs, SysArg, UNotifyEventRequest},
     sandbox::{Action, Capability, Sandbox, SandboxGuard},
+    syslog::LogLevel,
     warn,
 };
 
@@ -176,12 +197,15 @@ pub(crate) fn sandbox_path(
         return Err(if hide { Errno::ENOENT } else { Errno::EACCES });
     }
 
+    // Convert /proc/$pid to /proc/self as necessary.
+    let path = path.replace_proc_self(pid);
+
     // Check enabled capabilities.
     let caps_old = caps;
     let mut caps = sandbox.getcaps(caps);
     let stat = sandbox.enabled(Capability::CAP_STAT);
     if caps.is_empty() && (!hide || !stat) {
-        return if caps_old.can_write() && sandbox.is_append(path) {
+        return if caps_old.can_write() && sandbox.is_append(&path) {
             // SAFETY: Protect append-only paths against writes.
             // We use ECANCELED which will result in a no-op.
             Err(Errno::ECANCELED)
@@ -190,23 +214,7 @@ pub(crate) fn sandbox_path(
         };
     }
 
-    // Convert /proc/${pid} to /proc/self as necessary.
-    let path = if let Some(p) = path.split_prefix(b"/proc") {
-        let mut buf = itoa::Buffer::new();
-        let pid = buf.format(pid.as_raw());
-        if let Some(p) = p.split_prefix(pid.as_bytes()) {
-            let mut pdir = XPathBuf::from("/proc/self");
-            pdir.push(p.as_bytes());
-            Cow::Owned(pdir)
-        } else {
-            Cow::Borrowed(path)
-        }
-    } else {
-        Cow::Borrowed(path)
-    };
-
     let mut action = Action::Allow;
-    let mut filter = false;
     let check_caps = caps & Capability::CAP_PATH;
     let mut deny_errno = if check_caps == Capability::CAP_WALK {
         Errno::ENOENT
@@ -216,11 +224,10 @@ pub(crate) fn sandbox_path(
 
     // Sandboxing.
     for cap in check_caps {
-        let (new_action, new_filter) = sandbox.check_path(cap, &path);
+        let new_action = sandbox.check_path(cap, &path);
 
         if new_action >= action {
             action = new_action;
-            filter = new_filter;
         }
     }
 
@@ -228,17 +235,15 @@ pub(crate) fn sandbox_path(
     // for sandboxing types other than Stat.
     let check_hidden = stat && hide && (caps.is_empty() || action.is_denying());
     if check_hidden || caps.contains(Capability::CAP_STAT) {
-        let (new_action, new_filter) = sandbox.check_path(Capability::CAP_STAT, &path);
+        let new_action = sandbox.check_path(Capability::CAP_STAT, &path);
 
         if !check_hidden {
             deny_errno = Errno::ENOENT;
             action = new_action;
-            filter = new_filter;
         } else if new_action.is_denying() {
             deny_errno = Errno::ENOENT;
             if caps.is_empty() {
                 action = new_action;
-                filter = new_filter;
                 caps.insert(Capability::CAP_STAT);
             }
         }
@@ -249,7 +254,7 @@ pub(crate) fn sandbox_path(
         }
     }
 
-    if !filter && action >= Action::Warn {
+    if action.is_logging() && log_enabled!(LogLevel::Warn) {
         // Log warn for normal cases.
         // Log info for path hiding/walking unless explicitly specified to warn.
         let is_warn = match caps {
@@ -265,20 +270,11 @@ pub(crate) fn sandbox_path(
         };
 
         if let Some(request) = request {
-            let arch = request.scmpreq.data.arch;
             let args = request.scmpreq.data.args;
-
-            // Enrich log data with ioctl(2) request names.
-            let ctl = if caps.contains(Capability::CAP_IOCTL) {
-                ioctl_names_get(args[1] as Ioctl, arch)
-            } else {
-                None
-            };
-
-            if sandbox.verbose {
+            if sandbox.log_scmp() {
                 if is_warn {
                     warn!("ctx": "access", "cap": caps, "act": action,
-                        "sys": syscall_name, "ctl": ctl,
+                        "sys": syscall_name,
                         "path": &path, "args": args,
                         "tip": format!("configure `allow/{}+{}'",
                             caps.to_string().to_ascii_lowercase(),
@@ -286,7 +282,7 @@ pub(crate) fn sandbox_path(
                         "req": request);
                 } else {
                     notice!("ctx": "access", "cap": caps, "act": action,
-                        "sys": syscall_name, "ctl": ctl,
+                        "sys": syscall_name,
                         "path": &path, "args": args,
                         "tip": format!("configure `allow/{}+{}'",
                             caps.to_string().to_ascii_lowercase(),
@@ -295,7 +291,7 @@ pub(crate) fn sandbox_path(
                 }
             } else if is_warn {
                 warn!("ctx": "access", "cap": caps, "act": action,
-                    "sys": syscall_name, "ctl": ctl,
+                    "sys": syscall_name,
                     "path": &path, "args": args,
                     "tip": format!("configure `allow/{}+{}'",
                         caps.to_string().to_ascii_lowercase(),
@@ -303,7 +299,7 @@ pub(crate) fn sandbox_path(
                     "pid": request.scmpreq.pid);
             } else {
                 notice!("ctx": "access", "cap": caps, "act": action,
-                    "sys": syscall_name, "ctl": ctl,
+                    "sys": syscall_name,
                     "path": &path, "args": args,
                     "tip": format!("configure `allow/{}+{}'",
                         caps.to_string().to_ascii_lowercase(),
@@ -405,8 +401,8 @@ where
         //
         // Delay Chdir to allow the common `cd /` use case right after chroot(2).
         let sandbox = request.get_sandbox();
-        if sandbox.is_chroot() && !caps.contains(Capability::CAP_CHDIR) {
-            return Err(if is_fd { Errno::EACCES } else { Errno::ENOENT });
+        if sandbox.is_chroot() && !is_fd && !caps.contains(Capability::CAP_CHDIR) {
+            return Err(Errno::ENOENT);
         }
 
         // If sandboxing for all the selected capabilities is off, return immediately.
@@ -425,8 +421,8 @@ where
 
                 if sandbox.is_chroot() {
                     return if caps.contains(Capability::CAP_CHDIR) && path.abs().is_rootfs() {
-                        // SAFETY: Allow `cd /` after chroot.
-                        Ok(unsafe { request.continue_syscall() })
+                        // No-op `cd /` after chroot.
+                        Ok(request.return_syscall(0))
                     } else {
                         // arg.path.is_some() -> is_fd = false: Return ENOENT.
                         Err(Errno::ENOENT)
@@ -438,25 +434,15 @@ where
                 // Validate FD argument.
                 let dirfd = to_valid_fd(req.data.args[arg_idx])?;
 
-                if sandbox.is_chroot() {
-                    // Return EACCES for FD-only system calls such as fchdir.
-                    return Err(if is_fd { Errno::EACCES } else { Errno::ENOENT });
-                }
-
                 if dirfd != libc::AT_FDCWD {
                     // SAFETY: Get the file descriptor before access check
                     // as it may change after which is a TOCTOU vector.
                     let fd = request.get_fd(dirfd)?;
 
-                    // Validate WANT_READ against O_PATH.
-                    if arg.fsflags.want_read() && fd_status_flags(&fd)?.contains(OFlag::O_PATH) {
-                        return Err(Errno::EBADF);
-                    }
-
                     // Handle ftruncate etc. for files with encryption in progress.
                     let crypt_path = if crypt {
                         #[expect(clippy::disallowed_methods)]
-                        let files = request.crypt_map.as_ref().unwrap();
+                        let files = request.cache.crypt_map.as_ref().unwrap();
                         if let Ok(info) = FileInfo::from_fd(&fd) {
                             let files = files.0.lock().unwrap_or_else(|e| e.into_inner());
                             files
@@ -645,7 +631,7 @@ pub(crate) fn to_sflag(arg: u64) -> Result<SFlag, Errno> {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::fs::AT_EXECVE_CHECK;
+    use crate::fd::AT_EXECVE_CHECK;
 
     #[test]
     fn test_to_atflags() {
