@@ -1,28 +1,31 @@
 // Syd: rock-solid application kernel
 // src/kernel/net/recvfrom.rs: recvfrom(2) handler
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
-use std::os::{
-    fd::{AsFd, AsRawFd, OwnedFd},
-    unix::ffi::OsStrExt,
+use std::{
+    mem::MaybeUninit,
+    os::{
+        fd::{AsFd, AsRawFd, OwnedFd},
+        unix::ffi::OsStrExt,
+    },
 };
 
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
-    sys::socket::{recv, MsgFlags, SockaddrLike, SockaddrStorage},
+    sys::socket::{recv, SockaddrLike, SockaddrStorage},
 };
 use zeroize::Zeroizing;
 
 use crate::{
-    config::MMAP_MIN_ADDR,
-    fs::{get_nonblock, has_recv_timeout},
-    hook::UNotifyEventRequest,
+    compat::MsgFlags,
+    fd::{get_nonblock, has_recv_timeout},
     kernel::net::to_msgflags,
     path::XPath,
+    req::UNotifyEventRequest,
 };
 
 pub(crate) fn handle_recv(
@@ -70,7 +73,7 @@ pub(crate) fn handle_recv(
         false
     };
 
-    let result = recv(fd.as_raw_fd(), &mut buf, flags);
+    let result = recv(fd.as_raw_fd(), &mut buf, flags.into());
 
     // Remove invalidation record unless interrupted.
     if is_blocking {
@@ -99,13 +102,13 @@ pub(crate) fn handle_recvfrom(
     let addrlen = if args[5] != 0 {
         const SIZEOF_SOCKLEN_T: usize = size_of::<libc::socklen_t>();
         let mut buf = [0u8; SIZEOF_SOCKLEN_T];
-        if request.read_mem(&mut buf, args[5])? == SIZEOF_SOCKLEN_T {
+        if request.read_mem(&mut buf, args[5], SIZEOF_SOCKLEN_T)? == SIZEOF_SOCKLEN_T {
             // libc defines socklen_t as u32,
             // however we should check for negative values
             // and return EINVAL as necessary.
             let len = i32::from_ne_bytes(buf);
             let len = libc::socklen_t::try_from(len).or(Err(Errno::EINVAL))?;
-            if args[4] < *MMAP_MIN_ADDR {
+            if len > 0 && args[4] == 0 {
                 // address length is positive however address is NULL:
                 // Return EINVAL and NOT EFAULT here, see LTP accept01 check.
                 return Err(Errno::EINVAL);
@@ -164,6 +167,7 @@ pub(crate) fn handle_recvfrom(
 
     // Write address into sandbox process memory as necessary.
     // The address may be None for connection-mode sockets.
+    #[expect(clippy::cast_possible_truncation)]
     let len = if let Some((addr, addrlen_out)) = addr {
         // Create a byte slice from the socket address pointer.
         //
@@ -223,14 +227,12 @@ fn do_recvfrom<Fd: AsFd>(
         buf.resize(len, 0);
     }
 
-    // Allocate storage for the address.
-    let mut addr: [u8; std::mem::size_of::<SockaddrStorage>()] =
-        [0u8; std::mem::size_of::<SockaddrStorage>()];
+    // Allocate properly aligned storage for the address.
+    let mut addr = MaybeUninit::<SockaddrStorage>::zeroed();
     #[expect(clippy::cast_possible_truncation)]
-    let mut len = std::mem::size_of::<SockaddrStorage>() as libc::socklen_t;
+    let mut len = size_of::<SockaddrStorage>() as libc::socklen_t;
 
-    // Cast the storage buffer to a sockaddr pointer.
-    #[expect(clippy::cast_ptr_alignment)]
+    // Cast the aligned storage to a sockaddr pointer.
     let ptr = addr.as_mut_ptr() as *mut libc::sockaddr;
 
     // SAFETY: Record blocking call so it can get invalidated.
