@@ -1,7 +1,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/net/sendmsg.rs: sendmsg(2) and sendmmsg(2) handlers
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -20,18 +20,20 @@ use zeroize::Zeroizing;
 
 use crate::{
     compat::{
-        addr_family, cmsg_len_32, cmsg_space_32, msghdr, msghdr32, MsgFlags, PF_ALG, PF_INET,
-        PF_INET6, PF_NETLINK, PF_UNIX, UIO_MAXIOV, UIO_MAXLEN,
+        addr_family, cmsg_len_32, cmsg_space_32, getsockdomain, msghdr, msghdr32, MsgFlags, PF_ALG,
+        PF_INET, PF_INET6, PF_NETLINK, PF_UNIX, UIO_MAXIOV,
     },
+    config::MAX_RW_COUNT,
     confine::scmp_arch_bits,
-    fs::{get_nonblock, has_recv_timeout, is_same_vm},
+    fd::{get_nonblock, has_recv_timeout},
+    fs::is_same_vm,
     kernel::{
-        net::{canon_addr, get_addr, sandbox_addr, to_msgflags},
+        net::{canon_addr, get_addr, make_lo6addr, make_loaddr, sandbox_addr, to_msgflags},
         sandbox_path,
     },
     path::XPath,
     req::UNotifyEventRequest,
-    sandbox::{Capability, Flags},
+    sandbox::{Capability, Flags, Options},
 };
 
 #[expect(clippy::cognitive_complexity)]
@@ -40,12 +42,13 @@ pub(crate) fn handle_sendmsg(
     request: &UNotifyEventRequest,
     args: &[u64; 6],
     flags: Flags,
+    options: Options,
 ) -> Result<ScmpNotifResp, Errno> {
     // SAFETY: Reject undefined/invalid flags.
     let msgflags = to_msgflags(args[2])?;
 
     // SAFETY: Reject MSG_OOB as necessary.
-    if !flags.allow_unsafe_oob() && msgflags.contains(MsgFlags::MSG_OOB) {
+    if !options.allow_unsafe_oob() && msgflags.contains(MsgFlags::MSG_OOB) {
         // Signal no support to let the sandbox process
         // handle the error gracefully. This is consistent
         // with the Linux kernel.
@@ -85,12 +88,49 @@ pub(crate) fn handle_sendmsg(
 
     // Step 1: Handle the address.
     let sandbox = request.get_sandbox();
+    let any_addr = sandbox.flags.allow_unsafe_any_addr();
+    let local_net = sandbox.flags.force_local_net();
     let log_scmp = sandbox.log_scmp();
+    let mut addr_unix = false;
     let addr = if !addr_remote.is_null() && addr_len > 0 {
         let addr = get_addr(request, addr_remote as u64, addr_len)?;
-        let (addr, root) = canon_addr(request, &sandbox, &addr, Capability::CAP_NET_CONNECT)?;
+        let (mut addr, root) = canon_addr(request, &sandbox, &addr, Capability::CAP_NET_CONNECT)?;
         match addr_family(&addr) {
-            PF_UNIX | PF_INET | PF_INET6 => {
+            PF_UNIX => {
+                // Check for access.
+                sandbox_addr(
+                    request,
+                    &sandbox,
+                    &addr,
+                    &root,
+                    0x10,
+                    Capability::CAP_NET_CONNECT,
+                )?;
+                addr_unix = true;
+            }
+            PF_INET => {
+                // Turn 0.0.0.0 into 127.0.0.1 unless trace/allow_unsafe_any_addr:1.
+                // Turn all addresses to 127.0.0.1 if trace/force_local_net:1.
+                if !any_addr {
+                    make_loaddr(&mut addr, local_net, 0x10 /*sendmsg*/)?;
+                }
+
+                // Check for access.
+                sandbox_addr(
+                    request,
+                    &sandbox,
+                    &addr,
+                    &root,
+                    0x10,
+                    Capability::CAP_NET_CONNECT,
+                )?;
+            }
+            PF_INET6 => {
+                // Turn :: into :1 unless trace/allow_unsafe_any_addr:1.
+                // Turn all addresses to ::1 if trace/force_local_net:1.
+                if !any_addr {
+                    make_lo6addr(&mut addr, local_net, 0x10 /*sendmsg*/)?;
+                }
                 // Check for access.
                 sandbox_addr(
                     request,
@@ -104,7 +144,7 @@ pub(crate) fn handle_sendmsg(
             PF_ALG | PF_NETLINK => {
                 // SAFETY: We do not check AF_ALG or AF_NETLINK for access.
             }
-            _ if flags.allow_unsupp_socket() => {
+            _ if options.allow_unsupp_socket() => {
                 // SAFETY: We do not check unsupported sockets for access.
             }
             _ => return Err(Errno::EAFNOSUPPORT),
@@ -113,8 +153,13 @@ pub(crate) fn handle_sendmsg(
         Some((addr, root))
     } else {
         // Connection-mode socket.
-        // SAFETY: We cannot continue here due to the added level of
-        // pointer indirection.
+        // SAFETY:
+        // 1. We cannot continue here due to the added level of pointer indirection.
+        // 2. We need to check whether this is an AF_UNIX socket.
+        addr_unix = getsockdomain(&fd)
+            .map(|domain| domain == libc::AF_UNIX)
+            .unwrap_or(false);
+
         None
     };
 
@@ -124,13 +169,14 @@ pub(crate) fn handle_sendmsg(
         #[expect(clippy::useless_conversion)]
         let cmsg_len = usize::try_from(msg.msg_controllen)
             .or(Err(Errno::EINVAL))?
-            .min(UIO_MAXLEN); // SAFETY: Cap at UIO_MAXLEN.
+            .min(*MAX_RW_COUNT); // SAFETY: Cap at MAX_RW_COUNT.
         let cmsg_buf = request.read_vec(msg.msg_control as u64, cmsg_len)?;
         Some(parse_control_messages(request, &cmsg_buf)?)
     } else {
         None
     };
 
+    #[allow(unused)]
     if let Some((
         ref control_fds,
         ref control_creds,
@@ -203,6 +249,7 @@ pub(crate) fn handle_sendmsg(
             control_messages.push(ControlMessage::AlgSetAeadAssoclen(assoclen));
         }
 
+        #[cfg(not(target_os = "android"))]
         for gso_segments in control_udp_gso_segments {
             control_messages.push(ControlMessage::UdpGsoSegments(gso_segments));
         }
@@ -219,6 +266,7 @@ pub(crate) fn handle_sendmsg(
             control_messages.push(ControlMessage::RxqOvfl(rxq_ovfl));
         }
 
+        #[cfg(not(target_os = "android"))]
         for tx_time in control_tx_times {
             control_messages.push(ControlMessage::TxTime(tx_time));
         }
@@ -277,7 +325,7 @@ pub(crate) fn handle_sendmsg(
             // Validate length to avoid large allocations.
             #[expect(clippy::useless_conversion)]
             let iov_len = usize::try_from(iov.iov_len).or(Err(Errno::EMSGSIZE))?;
-            if iov_len > UIO_MAXLEN {
+            if iov_len > *MAX_RW_COUNT {
                 // Invalid iovec count: Too large.
                 return Err(Errno::EMSGSIZE);
             }
@@ -340,7 +388,13 @@ pub(crate) fn handle_sendmsg(
     // Send SIGPIPE for EPIPE unless MSG_NOSIGNAL is set.
     #[expect(clippy::cast_possible_wrap)]
     Ok(match result {
-        Ok(n) => request.return_syscall(n as i64),
+        Ok(n) => {
+            // Record sender PID for SCM_PIDFD/SO_PASSCRED fixup at recvmsg(2).
+            if addr_unix {
+                let _ = request.add_unix(&fd, req.pid(), None, None);
+            }
+            request.return_syscall(n as i64)
+        }
         Err(Errno::EPIPE) if !msgflags.contains(MsgFlags::MSG_NOSIGNAL) => {
             request.pidfd_kill(libc::SIGPIPE)?;
             request.fail_syscall(Errno::EPIPE)
@@ -355,12 +409,13 @@ pub(crate) fn handle_sendmmsg(
     request: &UNotifyEventRequest,
     args: &[u64; 6],
     flags: Flags,
+    options: Options,
 ) -> Result<ScmpNotifResp, Errno> {
     // SAFETY: Reject undefined/invalid flags.
     let msgflags = to_msgflags(args[3])?;
 
     // SAFETY: Reject MSG_OOB as necessary.
-    if !flags.allow_unsafe_oob() && msgflags.contains(MsgFlags::MSG_OOB) {
+    if !options.allow_unsafe_oob() && msgflags.contains(MsgFlags::MSG_OOB) {
         // Signal no support to let the sandbox process
         // handle the error gracefully. This is consistent
         // with the Linux kernel.
@@ -418,7 +473,12 @@ pub(crate) fn handle_sendmmsg(
     // Check if we want to send SIGPIPE on EPIPE.
     let must_signal = !msgflags.contains(MsgFlags::MSG_NOSIGNAL);
 
-    // Prepare a series of sendmsg calls.
+    // Prepare a series of sendmsg(2) calls.
+    // Send messages in order, break on error,
+    // return the number of successfully sent messages if > 0.
+    let mut addr_unix = false;
+    let mut datagrams: usize = 0;
+    let mut last_errno: Option<Errno> = None;
     for mmsg in &mut mmsghdrs {
         let msg = &mut mmsg.msg_hdr;
 
@@ -426,12 +486,48 @@ pub(crate) fn handle_sendmmsg(
         let addr_remote = msg.msg_name;
         let addr_len = msg.msg_namelen;
         let sandbox = request.get_sandbox();
+        let any_addr = sandbox.flags.allow_unsafe_any_addr();
+        let local_net = sandbox.flags.force_local_net();
         let log_scmp = sandbox.log_scmp();
         let addr_root = if !addr_remote.is_null() && addr_len > 0 {
             let addr = get_addr(request, addr_remote as u64, addr_len)?;
-            let (addr, root) = canon_addr(request, &sandbox, &addr, Capability::CAP_NET_CONNECT)?;
+            let (mut addr, root) =
+                canon_addr(request, &sandbox, &addr, Capability::CAP_NET_CONNECT)?;
             match addr_family(&addr) {
-                PF_UNIX | PF_INET | PF_INET6 => {
+                PF_UNIX => {
+                    // Check for access.
+                    sandbox_addr(
+                        request,
+                        &sandbox,
+                        &addr,
+                        &root,
+                        0x14,
+                        Capability::CAP_NET_CONNECT,
+                    )?;
+                    addr_unix = true;
+                }
+                PF_INET => {
+                    // Turn 0.0.0.0 into 127.0.0.1 unless trace/allow_unsafe_any_addr:1.
+                    // Turn all addresses to 127.0.0.1 if trace/force_local_net:1.
+                    if !any_addr {
+                        make_loaddr(&mut addr, local_net, 0x14 /*sendmmsg*/)?;
+                    }
+                    // Check for access.
+                    sandbox_addr(
+                        request,
+                        &sandbox,
+                        &addr,
+                        &root,
+                        0x14,
+                        Capability::CAP_NET_CONNECT,
+                    )?;
+                }
+                PF_INET6 => {
+                    // Turn :: into :1 unless trace/allow_unsafe_any_addr:1.
+                    // Turn all addresses to ::1 if trace/force_local_net:1.
+                    if !any_addr {
+                        make_lo6addr(&mut addr, local_net, 0x14 /*sendmmsg*/)?;
+                    }
                     // Check for access.
                     sandbox_addr(
                         request,
@@ -445,7 +541,7 @@ pub(crate) fn handle_sendmmsg(
                 PF_ALG | PF_NETLINK => {
                     // SAFETY: We do not check AF_ALG and AF_NETLINK for access.
                 }
-                _ if flags.allow_unsupp_socket() => {
+                _ if options.allow_unsupp_socket() => {
                     // SAFETY: We do not check unsupported sockets for access.
                 }
                 _ => return Err(Errno::EAFNOSUPPORT),
@@ -454,8 +550,15 @@ pub(crate) fn handle_sendmmsg(
             Some((addr, root))
         } else {
             // Connection-mode socket.
-            // SAFETY: We cannot continue here due to the added level of
-            // pointer indirection.
+            // SAFETY:
+            // 1. We cannot continue here due to the added level of pointer indirection.
+            // 2. We need to check whether this is an AF_UNIX socket.
+            if !addr_unix {
+                addr_unix = getsockdomain(&fd)
+                    .map(|domain| domain == libc::AF_UNIX)
+                    .unwrap_or(false);
+            }
+
             None
         };
 
@@ -465,7 +568,7 @@ pub(crate) fn handle_sendmmsg(
             #[expect(clippy::useless_conversion)]
             let cmsg_len = usize::try_from(msg.msg_controllen)
                 .or(Err(Errno::EINVAL))?
-                .min(UIO_MAXLEN); // SAFETY: Cap at UIO_MAXLEN.
+                .min(*MAX_RW_COUNT); // SAFETY: Cap at MAX_RW_COUNT.
             let cmsg_buf = request.read_vec(msg.msg_control as u64, cmsg_len)?;
             Some(parse_control_messages(request, &cmsg_buf)?)
         } else {
@@ -476,6 +579,7 @@ pub(crate) fn handle_sendmmsg(
 
         let mut control_messages = Vec::new();
         for control_datum in &control_data {
+            #[allow(unused)]
             if let Some((
                 ref control_fds,
                 ref control_creds,
@@ -549,6 +653,7 @@ pub(crate) fn handle_sendmmsg(
                     control_messages.push(ControlMessage::AlgSetAeadAssoclen(assoclen));
                 }
 
+                #[cfg(not(target_os = "android"))]
                 for gso_segments in control_udp_gso_segments {
                     control_messages.push(ControlMessage::UdpGsoSegments(gso_segments));
                 }
@@ -565,6 +670,7 @@ pub(crate) fn handle_sendmmsg(
                     control_messages.push(ControlMessage::RxqOvfl(rxq_ovfl));
                 }
 
+                #[cfg(not(target_os = "android"))]
                 for tx_time in control_tx_times {
                     control_messages.push(ControlMessage::TxTime(tx_time));
                 }
@@ -626,7 +732,7 @@ pub(crate) fn handle_sendmmsg(
                 // Validate length to avoid large allocations.
                 #[expect(clippy::useless_conversion)]
                 let iov_len = usize::try_from(iov.iov_len).or(Err(Errno::EMSGSIZE))?;
-                if iov_len > UIO_MAXLEN {
+                if iov_len > *MAX_RW_COUNT {
                     // Invalid iovec count: Too large.
                     return Err(Errno::EMSGSIZE);
                 }
@@ -677,23 +783,27 @@ pub(crate) fn handle_sendmmsg(
             }
 
             // Send SIGPIPE for EPIPE unless MSG_NOSIGNAL is set.
-            mmsg.msg_len = match result {
-                Ok(n) => n.try_into().or(Err(Errno::EINVAL))?,
+            match result {
+                Ok(n) => mmsg.msg_len = n.try_into().or(Err(Errno::EINVAL))?,
                 Err(Errno::EPIPE) if must_signal => {
-                    request.pidfd_kill(libc::SIGPIPE)?;
-                    return Err(Errno::EPIPE);
+                    let _ = request.pidfd_kill(libc::SIGPIPE);
+                    last_errno = Some(Errno::EPIPE);
+                    break;
+                }
+                Err(errno) => {
+                    last_errno = Some(errno);
+                    break;
                 }
-                Err(errno) => return Err(errno),
             };
-
-            mmsg.msg_len = result?.try_into().or(Err(Errno::EINVAL))?;
         }
+
+        datagrams = datagrams.checked_add(1).ok_or(Errno::EOVERFLOW)?;
     }
 
-    // Write back mmsghdr structures to remote process memory
+    // Write back mmsghdr structures for successfully sent messages.
     let mut buf: Zeroizing<Vec<u8>> = Zeroizing::new(Vec::new());
     if is32 {
-        for mmsghdr in &mmsghdrs {
+        for mmsghdr in &mmsghdrs[..datagrams] {
             let mmsghdr32: crate::compat::mmsghdr32 = (*mmsghdr).into();
             // SAFETY: Convert each mmsghdr (or mmsghdr32 within the
             // conversion logic) back to its byte representation.
@@ -703,7 +813,7 @@ pub(crate) fn handle_sendmmsg(
             buf.extend_from_slice(&bytes);
         }
     } else {
-        for mmsghdr in &mmsghdrs {
+        for mmsghdr in &mmsghdrs[..datagrams] {
             // SAFETY: See above.
             let bytes: [u8; size_of::<crate::compat::mmsghdr>()] =
                 unsafe { std::mem::transmute(*mmsghdr) };
@@ -711,11 +821,23 @@ pub(crate) fn handle_sendmmsg(
             buf.extend_from_slice(&bytes);
         }
     }
-    request.write_mem(&buf, addr)?;
+    if !buf.is_empty() {
+        request.write_mem(&buf, addr)?;
+    }
 
-    // FIXME: We do not handle partial success.
-    #[expect(clippy::cast_possible_wrap)]
-    Ok(request.return_syscall(mmsghdrs.len() as i64))
+    // Record sender PID for SCM_PIDFD/SO_PASSCRED fixup at recvmsg(2).
+    if addr_unix {
+        let _ = request.add_unix(&fd, req.pid(), None, None);
+    }
+
+    // Return the number of successfully sent messages if at least one succeeded,
+    // otherwise return the error from the first failed message.
+    if datagrams > 0 {
+        #[expect(clippy::cast_possible_wrap)]
+        Ok(request.return_syscall(datagrams as i64))
+    } else {
+        Err(last_errno.unwrap_or(Errno::EIO))
+    }
 }
 
 // SAFETY: Below lie daemons...
@@ -745,10 +867,12 @@ fn parse_control_messages(
     let mut control_ivs = Vec::new();
     let mut control_ops = Vec::new();
     let mut control_aead_assoclens = Vec::new();
+    #[allow(unused)]
     let mut control_udp_gso_segments = Vec::new();
     let mut control_ipv4_packet_infos = Vec::new();
     let mut control_ipv6_packet_infos = Vec::new();
     let mut control_rxq_ovfls = Vec::new();
+    #[allow(unused)]
     let mut control_tx_times = Vec::new();
     let mut control_ipv4_toses = Vec::new();
     let mut control_ipv6_tclasses = Vec::new();
@@ -821,29 +945,23 @@ fn parse_control_messages(
                 for i in 0..fd_count {
                     #[expect(clippy::arithmetic_side_effects)]
                     let fd_offset = i * size_of::<RawFd>();
-                    #[expect(clippy::cast_ptr_alignment)]
-                    // SAFETY: See the comment above.
-                    let fd = unsafe { *(data[fd_offset..].as_ptr() as *const RawFd) };
+                    // SAFETY: Unaligned read of RawFd from u8 cmsg buffer.
+                    let fd = unsafe {
+                        std::ptr::read_unaligned(data[fd_offset..].as_ptr() as *const RawFd)
+                    };
                     let fd = request.get_fd(fd)?;
                     fds.push(fd);
                 }
                 control_fds.push(fds);
             }
             (libc::SOL_SOCKET, libc::SCM_CREDENTIALS) => {
-                #[expect(clippy::cast_ptr_alignment)]
-                // SAFETY: Casting `data.as_ptr()` to `*const libc::ucred` is safe
-                // under the assumption that `data` contains bytes that correctly represent
-                // a `libc::ucred` structure, and that `data_len` matches the size of
-                // `libc::ucred`.  This assumption is based on the control message type
-                // `SCM_CREDENTIALS`, which is expected to contain exactly one `libc::ucred`
-                // structure representing the credentials of the sending process.  The
-                // conversion to `UnixCredentials` is a safe operation that simply wraps the
-                // raw credentials in a Rust-friendly type. The use of `unsafe` is necessary
-                // to dereference the raw pointer obtained from the byte buffer, but the
-                // operation is ensured to be valid by adhering to the expected control
-                // message format and size. This access pattern is consistent with the
-                // standard way of handling `SCM_CREDENTIALS` control messages.
-                let mut creds = *(unsafe { &*(data.as_ptr() as *const libc::ucred) });
+                if data_len != size_of::<libc::ucred>() {
+                    return Err(Errno::EINVAL); // Data length mismatch
+                }
+                // SAFETY: Unaligned read of POD `libc::ucred` from u8 cmsg buffer;
+                // data_len validated above.
+                let mut creds: libc::ucred =
+                    unsafe { std::ptr::read_unaligned(data.as_ptr() as *const libc::ucred) };
                 // SAFETY: The sender must specify its own pid (unless it has the capability
                 // CAP_SYS_ADMIN, in which case the PID of any existing process may be specified.)
                 if creds.pid != request.scmpreq.pid().as_raw() {
@@ -892,9 +1010,9 @@ fn parse_control_messages(
                 if data_len != size_of::<libc::c_int>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[expect(clippy::cast_ptr_alignment)]
-                // SAFETY: See the comment above.
-                let op = unsafe { *(data.as_ptr() as *const libc::c_int) };
+                // SAFETY: Unaligned read of c_int from u8 cmsg buffer;
+                // data_len validated above.
+                let op = unsafe { std::ptr::read_unaligned(data.as_ptr() as *const libc::c_int) };
                 control_ops.push(op); // Store the operation code
             }
             (libc::SOL_ALG, libc::ALG_SET_AEAD_ASSOCLEN) => {
@@ -913,11 +1031,12 @@ fn parse_control_messages(
                 if data_len != size_of::<u32>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[expect(clippy::cast_ptr_alignment)]
-                // SAFETY: See the comment above.
-                let assoclen = unsafe { *(data.as_ptr() as *const u32) };
+                // SAFETY: Unaligned read of u32 from u8 cmsg buffer;
+                // data_len validated above.
+                let assoclen = unsafe { std::ptr::read_unaligned(data.as_ptr() as *const u32) };
                 control_aead_assoclens.push(assoclen); // Store the AEAD assoclen
             }
+            #[cfg(not(target_os = "android"))]
             (libc::SOL_UDP, libc::UDP_SEGMENT) => {
                 // SAFETY: This unsafe block casts a pointer from `*const u8` to `*const
                 // u16` under the assumption that the data at `data.as_ptr()` is correctly
@@ -934,9 +1053,9 @@ fn parse_control_messages(
                 if data_len != size_of::<u16>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[expect(clippy::cast_ptr_alignment)]
-                // SAFETY: See the comment above.
-                let gso_segments = unsafe { *(data.as_ptr() as *const u16) };
+                // SAFETY: Unaligned read of u16 from u8 cmsg buffer;
+                // data_len validated above.
+                let gso_segments = unsafe { std::ptr::read_unaligned(data.as_ptr() as *const u16) };
                 control_udp_gso_segments.push(gso_segments); // Store the GSO segment count
             }
             (libc::IPPROTO_IP, libc::IP_PKTINFO) => {
@@ -954,10 +1073,11 @@ fn parse_control_messages(
                 if data_len != size_of::<libc::in_pktinfo>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[expect(clippy::cast_ptr_alignment)]
-                // SAFETY: See the comment above.
-                let pktinfo = unsafe { &*(data.as_ptr() as *const libc::in_pktinfo) };
-                control_ipv4_packet_infos.push(*pktinfo); // Store the IPv4 packet info
+                // SAFETY: Unaligned read of in_pktinfo from u8 cmsg buffer;
+                // data_len validated above.
+                let pktinfo: libc::in_pktinfo =
+                    unsafe { std::ptr::read_unaligned(data.as_ptr() as *const libc::in_pktinfo) };
+                control_ipv4_packet_infos.push(pktinfo); // Store the IPv4 packet info
             }
             (libc::IPPROTO_IPV6, libc::IPV6_PKTINFO) => {
                 // SAFETY: The cast from `*const u8` to `*const libc::in6_pktinfo` is made
@@ -974,10 +1094,11 @@ fn parse_control_messages(
                 if data_len != size_of::<libc::in6_pktinfo>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[expect(clippy::cast_ptr_alignment)]
-                // SAFETY: See the comment above.
-                let pktinfo = unsafe { &*(data.as_ptr() as *const libc::in6_pktinfo) };
-                control_ipv6_packet_infos.push(*pktinfo); // Store the IPv6 packet info
+                // SAFETY: Unaligned read of in6_pktinfo from u8 cmsg buffer;
+                // data_len validated above.
+                let pktinfo: libc::in6_pktinfo =
+                    unsafe { std::ptr::read_unaligned(data.as_ptr() as *const libc::in6_pktinfo) };
+                control_ipv6_packet_infos.push(pktinfo); // Store the IPv6 packet info
             }
             (libc::SOL_SOCKET, libc::SO_RXQ_OVFL) => {
                 // SAFETY: Casting from `*const u8` to `*const u32` here assumes that the
@@ -995,11 +1116,12 @@ fn parse_control_messages(
                 if data_len != size_of::<u32>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[expect(clippy::cast_ptr_alignment)]
-                // SAFETY: See the comment above.
-                let rxq_ovfl = unsafe { *(data.as_ptr() as *const u32) };
+                // SAFETY: Unaligned read of u32 from u8 cmsg buffer;
+                // data_len validated above.
+                let rxq_ovfl = unsafe { std::ptr::read_unaligned(data.as_ptr() as *const u32) };
                 control_rxq_ovfls.push(rxq_ovfl); // Store the Rx queue overflow count
             }
+            #[cfg(not(target_os = "android"))]
             (libc::SOL_SOCKET, libc::SCM_TXTIME) => {
                 // SAFETY: The casting from `*const u8` to `*const u64` is contingent upon
                 // the assumption that the `data` buffer is aligned according to `u64`
@@ -1015,9 +1137,9 @@ fn parse_control_messages(
                 if data_len != size_of::<u64>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[expect(clippy::cast_ptr_alignment)]
-                // SAFETY: See the comment above.
-                let tx_time = unsafe { *(data.as_ptr() as *const u64) };
+                // SAFETY: Unaligned read of u64 from u8 cmsg buffer;
+                // data_len validated above.
+                let tx_time = unsafe { std::ptr::read_unaligned(data.as_ptr() as *const u64) };
                 control_tx_times.push(tx_time); // Store the Tx time
             }
             (libc::IPPROTO_IP, libc::IP_TOS) => {
@@ -1034,10 +1156,9 @@ fn parse_control_messages(
                 if data_len != size_of::<i32>() {
                     return Err(Errno::EINVAL);
                 }
-                // SAFETY: Exact int-sized payload validated; bounds and alignment guaranteed
-                // by prior checks and cmsg parsing semantics. Single c_int read is safe.
-                #[expect(clippy::cast_ptr_alignment)]
-                let val = unsafe { *(data.as_ptr() as *const i32) };
+                // SAFETY: Unaligned read of i32 from u8 cmsg buffer;
+                // data_len validated above.
+                let val = unsafe { std::ptr::read_unaligned(data.as_ptr() as *const i32) };
                 control_ipv6_tclasses.push(val);
             }
             _ => return Err(Errno::EINVAL),
