@@ -1,7 +1,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/net/sendto.rs: sendto(2) handler
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -10,13 +10,13 @@ use std::os::fd::{AsRawFd, OwnedFd};
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
-    sys::socket::{send, sendto, SockaddrStorage},
+    sys::socket::{send, sendto, AddressFamily, SockaddrLike, SockaddrStorage},
 };
-use zeroize::Zeroizing;
 
 use crate::{
-    compat::MsgFlags,
-    fs::{get_nonblock, has_recv_timeout},
+    compat::{getsockdomain, MsgFlags},
+    config::MAX_RW_COUNT,
+    fd::{get_nonblock, has_recv_timeout},
     kernel::net::to_msgflags,
     req::UNotifyEventRequest,
 };
@@ -47,13 +47,11 @@ pub(crate) fn handle_sendto(
     // 3. Buffer read from sandbox process MUST be zeroized on drop.
     let len = usize::try_from(args[2])
         .or(Err(Errno::EINVAL))?
-        .min(1000000); // Cap count at 1mio.
-    let mut buf = Zeroizing::new(Vec::new());
-    if len > 0 {
-        buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
-        request.read_mem(&mut buf, args[1])?;
-    }
+        .min(*MAX_RW_COUNT); // Cap count at MAX_RW_COUNT.
+
+    // read_vec_zeroed returns an empty Vec with len==0,
+    // without performing any memory reads.
+    let buf = request.read_vec_zeroed(args[1], len)?;
 
     // SAFETY: Record blocking call so it can get invalidated.
     let req = request.scmpreq;
@@ -86,7 +84,21 @@ pub(crate) fn handle_sendto(
     // Send SIGPIPE for EPIPE unless MSG_NOSIGNAL is set.
     #[expect(clippy::cast_possible_wrap)]
     Ok(match result {
-        Ok(n) => request.return_syscall(n as i64),
+        Ok(n) => {
+            // Record sender PID for SCM_PIDFD/SO_PASSCRED fixup at recvmsg(2).
+            if addr
+                .as_ref()
+                .map(|addr| addr.family() == Some(AddressFamily::Unix))
+                .unwrap_or_else(|| {
+                    getsockdomain(&fd)
+                        .map(|domain| domain == libc::AF_UNIX)
+                        .unwrap_or(false)
+                })
+            {
+                let _ = request.add_unix(&fd, req.pid(), None, None);
+            }
+            request.return_syscall(n as i64)
+        }
         Err(Errno::EPIPE) if !flags.contains(MsgFlags::MSG_NOSIGNAL) => {
             request.pidfd_kill(libc::SIGPIPE)?;
             request.fail_syscall(Errno::EPIPE)
