@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/stat.rs: stat syscall handlers
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -16,7 +16,7 @@ use crate::{
         fstatat64, statx, FileStat64, FileStatx, FileStatxTimestamp, STATX_BASIC_STATS, STATX_MODE,
         STATX_TYPE,
     },
-    config::{API_VERSION, MAGIC_PREFIX},
+    config::{API_VERSION, MAGIC_LOAD, MAGIC_PREFIX},
     confine::{is_valid_ptr, scmp_arch_bits, EOWNERDEAD},
     fd::parse_fd,
     hash::SydHashSet,
@@ -24,7 +24,7 @@ use crate::{
     lookup::{CanonicalPath, FileInfo, FileType, FsFlags},
     path::XPath,
     req::{SysArg, SysFlags, UNotifyEventRequest},
-    sandbox::{Capability, Flags},
+    sandbox::{Capability, Options},
 };
 
 pub(crate) fn sys_stat(request: UNotifyEventRequest) -> ScmpNotifResp {
@@ -249,75 +249,22 @@ fn syscall_stat_handler(
         let restrict_stat_cdev = !sandbox.flags.allow_unsafe_stat_cdev();
         let mut ghost = false;
         let caps = *sandbox.state;
-        let sflags = *sandbox.flags;
+        let opts = *sandbox.options;
         if magic {
-            if sandbox.locked_for(req.pid()) {
+            if sandbox.locked_drop_for(req.pid()) {
                 // Sandbox is locked, access denied.
                 return Err(Errno::ENOENT);
             }
             drop(sandbox); // release the read-lock.
 
-            // Handle magic prefix (ie /dev/syd)
-            let mut cmd = path
+            // Handle magic prefix "/dev/syd".
+            let cmd = path
                 .abs()
                 .strip_prefix(MAGIC_PREFIX)
-                .unwrap_or_else(|| XPath::from_bytes(&path.abs().as_bytes()[MAGIC_PREFIX.len()..]))
-                .to_owned();
-            // Careful here, Path::strip_prefix removes trailing slashes.
-            if path.abs().ends_with_slash() {
-                cmd.push(b"");
-            }
+                .unwrap_or_else(|| XPath::from_bytes(&path.abs().as_bytes()[MAGIC_PREFIX.len()..]));
 
-            // Acquire a write lock to the sandbox.
-            let mut sandbox = request.get_mut_sandbox();
-
-            // Execute magic command.
-            if cmd.is_empty() || cmd.is_equal(b".el") || cmd.is_equal(b".sh") {
-                sandbox.config("")?;
-            } else if cmd.is_equal(b"panic") {
-                sandbox.panic()?;
-            } else if cmd.is_equal(b"ghost") {
-                // SAFETY: Reset sandbox to ensure no run-away execs.
-                sandbox.reset()?;
-                ghost = true;
-            } else if let Some(cmd) = cmd.strip_prefix(b"load") {
-                // We handle load specially here as it involves process access.
-                // 1. Attempt to parse as FD, pidfd_getfd and load it.
-                // 2. Attempt to parse as profile name if (1) fails.
-                match parse_fd(cmd) {
-                    Ok(remote_fd) => {
-                        let fd = request.get_fd(remote_fd)?;
-                        let file = BufReader::new(File::from(fd));
-                        let mut imap = SydHashSet::default();
-                        // SAFETY: parse_config() checks for the file name
-                        // /dev/syd/load and disables config file include
-                        // feature depending on this check.
-                        if sandbox
-                            .parse_config(file, XPath::from_bytes(b"/dev/syd/load"), &mut imap)
-                            .is_err()
-                        {
-                            return Ok(request.fail_syscall(Errno::EINVAL));
-                        }
-                        // Fall through to emulate as /dev/null.
-                    }
-                    Err(Errno::EBADF) => {
-                        if sandbox.parse_profile(&cmd.to_string()).is_err() {
-                            return Ok(request.fail_syscall(Errno::EINVAL));
-                        }
-                        // Fall through to emulate as /dev/null.
-                    }
-                    Err(errno) => {
-                        return Ok(request.fail_syscall(errno));
-                    }
-                }
-            } else if let Ok(cmd) = std::str::from_utf8(cmd.as_bytes()) {
-                sandbox.config(cmd)?;
-            } else {
-                // SAFETY: Invalid UTF-8 is not permitted.
-                // To include non-UTF-8, hex-encode them.
-                return Err(Errno::EINVAL);
-            }
-            drop(sandbox); // release the write-lock.
+            // Handle magic command.
+            ghost = handle_magic_stat(&request, cmd)?;
         } else {
             // Handle fstat for files with encryption in progress.
             #[expect(clippy::disallowed_methods)]
@@ -450,7 +397,7 @@ fn syscall_stat_handler(
 
             // If magic path, mask values for easy id.
             if magic {
-                magic_statx(&mut statx, caps, sflags);
+                magic_statx(&mut statx, caps, opts);
             }
 
             // SAFETY: The following block creates an immutable byte
@@ -497,7 +444,7 @@ fn syscall_stat_handler(
 
             // If magic path, mask values for easy id.
             if magic {
-                magic_stat(&mut stat, caps, sflags);
+                magic_stat(&mut stat, caps, opts);
             }
 
             let addr = req.data.args[arg_stat];
@@ -556,20 +503,83 @@ fn syscall_stat_handler(
     })
 }
 
-fn magic_stat(stat: &mut FileStat64, caps: Capability, flags: Flags) {
+fn handle_magic_stat(request: &UNotifyEventRequest, cmd: &XPath) -> Result<bool, Errno> {
+    // Set to true if ghost mode is initiated.
+    let mut ghost = false;
+
+    // Acquire a write lock to the sandbox.
+    let mut sandbox = request.get_mut_sandbox();
+
+    // Execute magic command.
+    if cmd.is_empty() || cmd.is_equal(b".el") || cmd.is_equal(b".sh") {
+        // Call no-ops into sandbox for logging as necessary.
+        sandbox.config("")?;
+    } else if cmd.is_equal(b"panic") {
+        // Rejects if locked or drop-only.
+        sandbox.panic()?;
+    } else if cmd.is_equal(b"ghost") {
+        // SAFETY:
+        // Reset sandbox to ensure no run-away execs:
+        // 1. Reset rejects if drop-only for !ghost.
+        // 2. Reset doesn't change state of Crypt sandboxing.
+        sandbox.reset(true)?;
+        ghost = true;
+    } else if let Some(cmd) = cmd.strip_prefix(b"load") {
+        // We handle load specially here as it involves process access.
+        // 1. Attempt to parse as FD, pidfd_getfd and load it.
+        // 2. Attempt to parse as profile name if (1) fails.
+        match parse_fd(cmd) {
+            Ok(remote_fd) => {
+                // SAFETY:
+                // parse_config() checks for the file name "/dev/syd/load", and
+                // disables config file include feature depending on this check.
+                let name = XPath::from_bytes(MAGIC_LOAD);
+                let file = request
+                    .get_fd(remote_fd)
+                    .map(File::from)
+                    .map(BufReader::new)?;
+                // FIXME: Do not waste getrandom(2) cycles for the unused hashset.
+                sandbox
+                    .parse_config(file, name, &mut SydHashSet::default() /*unused*/)
+                    .map_err(|err| err.errno().unwrap_or(Errno::EINVAL))?;
+                // Fall through to emulate as /dev/null.
+            }
+            Err(Errno::EBADF) => {
+                sandbox
+                    .parse_profile(cmd.as_bytes())
+                    .map_err(|err| err.errno().unwrap_or(Errno::EINVAL))?;
+                // Fall through to emulate as /dev/null.
+            }
+            Err(errno) => return Err(errno),
+        }
+    } else {
+        // SAFETY:
+        // Invalid UTF-8 is not permitted.
+        // To include non-UTF-8, user must hex-encode arguments.
+        std::str::from_utf8(cmd.as_bytes())
+            .or(Err(Errno::EINVAL))
+            .and_then(|cmd| sandbox.config(cmd))?;
+    }
+
+    // Sandbox write lock released here.
+    // Return true if ghost mode was initiated.
+    Ok(ghost)
+}
+
+fn magic_stat(stat: &mut FileStat64, caps: Capability, opts: Options) {
     stat.st_ino = 0;
     stat.st_nlink = caps.nlink().into();
-    stat.st_mode = magic_mode(caps, flags).into();
+    stat.st_mode = magic_mode(caps, opts).into();
     stat.st_rdev = API_VERSION.dev();
     stat.st_atime = 505958400; // must match libsyd!
     stat.st_ctime = -2036448000; // ditto!
     stat.st_mtime = -842745600; // ditto!
 }
 
-fn magic_statx(statx: &mut FileStatx, caps: Capability, flags: Flags) {
+fn magic_statx(statx: &mut FileStatx, caps: Capability, opts: Options) {
     statx.stx_ino = 0;
     statx.stx_nlink = caps.nlink();
-    statx.stx_mode = magic_mode(caps, flags);
+    statx.stx_mode = magic_mode(caps, opts);
     statx.stx_rdev_major = API_VERSION.major().into();
     statx.stx_rdev_minor = API_VERSION.minor().into();
     statx.stx_atime = FileStatxTimestamp {
@@ -588,18 +598,18 @@ fn magic_statx(statx: &mut FileStatx, caps: Capability, flags: Flags) {
 
 #[allow(clippy::cognitive_complexity)]
 #[allow(clippy::cast_possible_truncation)]
-fn magic_mode(caps: Capability, flags: Flags) -> u16 {
+fn magic_mode(caps: Capability, opts: Options) -> u16 {
     // Start with file type = character device
     let mut mode: u16 = libc::S_IFCHR as u16;
 
     // Special bits
-    if flags.contains(Flags::FL_UNSHARE_MOUNT) {
+    if opts.contains(Options::OPT_UNSHARE_MOUNT) {
         mode |= libc::S_ISVTX as u16;
     }
-    if flags.contains(Flags::FL_UNSHARE_USER) {
+    if opts.contains(Options::OPT_UNSHARE_USER) {
         mode |= libc::S_ISUID as u16;
     }
-    if flags.contains(Flags::FL_UNSHARE_NET) {
+    if opts.contains(Options::OPT_UNSHARE_NET) {
         mode |= libc::S_ISGID as u16;
     }
 
