@@ -2,12 +2,10 @@
 // Syd: rock-solid application kernel
 // src/kernel/unlink.rs: rmdir(2), unlink(2) and unlinkat(2) handlers
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
-use std::os::fd::AsFd;
-
 use libseccomp::ScmpNotifResp;
 use nix::{errno::Errno, unistd::UnlinkatFlags, NixPath};
 
@@ -88,7 +86,6 @@ fn syscall_unlink_handler(
     // SAFETY: SysArg has one element.
     #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
-    let dir = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
 
     // SAFETY: Path hiding is done, not it is safe to:
     // Return E{IS,NOT}DIR if path is not an expected dir/non-dir.
@@ -111,5 +108,5 @@ fn syscall_unlink_handler(
     } else {
         UnlinkatFlags::NoRemoveDir
     };
-    safe_unlinkat(dir, path.base(), flags).map(|_| request.return_syscall(0))
+    safe_unlinkat(path.dir(), path.base(), flags).map(|_| request.return_syscall(0))
 }
