@@ -70,9 +70,10 @@
 
 // We like clean and simple code with documentation.
 // Keep in sync with syd.rs.
+#![forbid(clippy::as_ptr_cast_mut)]
+#![forbid(clippy::cast_ptr_alignment)]
 #![deny(missing_docs)]
 #![deny(clippy::arithmetic_side_effects)]
-#![deny(clippy::as_ptr_cast_mut)]
 #![deny(clippy::as_underscore)]
 #![deny(clippy::assertions_on_result_states)]
 #![deny(clippy::borrow_as_ptr)]
@@ -82,7 +83,6 @@
 #![deny(clippy::cast_possible_truncation)]
 #![deny(clippy::cast_possible_wrap)]
 #![deny(clippy::cast_precision_loss)]
-#![deny(clippy::cast_ptr_alignment)]
 #![deny(clippy::cast_sign_loss)]
 #![deny(clippy::checked_conversions)]
 #![deny(clippy::clear_with_drain)]
@@ -111,6 +111,8 @@ pub mod api;
 /// Assembly instruction decoder
 #[cfg(feature = "asm")]
 pub mod asm;
+/// Binary entry points for multicall support
+pub mod bins;
 /// Cgroup v2 management for resource limits
 pub mod cgroup;
 /// System call argument cookies
@@ -234,6 +236,47 @@ pub(crate) mod pool;
 /// The low-level interface for linux namespaces (containers)
 pub mod unshare;
 
+/// KCOV userspace ABI shim for syzkaller
+#[cfg(feature = "kcov")]
+#[allow(clippy::arithmetic_side_effects)]
+#[allow(clippy::cast_possible_truncation)]
+#[allow(clippy::cast_possible_wrap)]
+#[allow(clippy::cast_sign_loss)]
+#[allow(clippy::disallowed_methods)]
+pub(crate) mod kcov;
+
+// Compat macros for instrumentation.
+// See kcov/api.rs for the implementations.
+#[cfg(not(feature = "kcov"))]
+#[allow(missing_docs)]
+#[macro_export]
+macro_rules! kcov_edge {
+    () => {{}};
+    ($site:expr) => {{}};
+}
+
+#[cfg(not(feature = "kcov"))]
+#[allow(missing_docs)]
+#[macro_export]
+macro_rules! kcov_edge_site {
+    ($s:literal) => {{}};
+}
+
+#[cfg(not(feature = "kcov"))]
+#[allow(missing_docs)]
+#[macro_export]
+macro_rules! kcov_cmp {
+    ($sz:expr, $isconst:expr, $a:expr, $b:expr) => {{}};
+    ($sz:expr, $isconst:expr, $a:expr, $b:expr, $site:expr) => {{}};
+}
+
+#[cfg(not(feature = "kcov"))]
+#[allow(missing_docs)]
+#[macro_export]
+macro_rules! kcov_cmp_site {
+    ($sz:expr, $isconst:expr, $a:expr, $b:expr, $s:literal) => {{}};
+}
+
 use std::ffi::OsStr;
 
 use lexis::ToName;
@@ -271,9 +314,29 @@ use crate::{
 /// Convenience macro to define a main function with correct errno return.
 #[macro_export]
 macro_rules! main {
+    { $name:ident => $($body:tt)* } => {
+        // Binary entry point for multicall support:
+        // This does not print to stderr(2) because self-confinement of
+        // utilities typically prevents it.
+        #[doc = concat!("Binary entry point: `", stringify!($name), "`.")]
+        pub fn $name() -> std::process::ExitCode {
+            match (|| -> $crate::err::SydResult<std::process::ExitCode> { $($body)* })() {
+                Ok(code) => code,
+                Err(err) => {
+                    u8::try_from(
+                        err.errno()
+                           .map(|e| e as i32)
+                           .unwrap_or(128)
+                    )
+                    .map(std::process::ExitCode::from)
+                    .unwrap_or(std::process::ExitCode::FAILURE)
+                }
+            }
+        }
+    };
     { $($body:tt)* } => {
         fn main() -> std::process::ExitCode {
-            match (|| -> syd::err::SydResult<std::process::ExitCode> { $($body)* })() {
+            match (|| -> $crate::err::SydResult<std::process::ExitCode> { $($body)* })() {
                 Ok(code) => code,
                 Err(err) => {
                     use std::io::Write;
@@ -554,6 +617,10 @@ pub fn syd_info(verbose: bool) -> Result<(), Errno> {
         ", {}pidfd_thread",
         if *HAVE_PIDFD_THREAD { "+" } else { "-" }
     )?;
+    printf!(
+        ", {}pidfd_get_info",
+        if *HAVE_PIDFD_GET_INFO { "+" } else { "-" }
+    )?;
     printf!(
         ", {}procmap_query",
         if *HAVE_PROCMAP_QUERY { "+" } else { "-" }
