@@ -46,7 +46,7 @@ use crate::{
     log::log_untrusted_buf,
     lookup::FileType,
     retry::retry_on_eintr,
-    sandbox::Flags,
+    sandbox::{Flags, Options},
 };
 
 /// Generate a formatted `XPathBuf`.
@@ -440,13 +440,13 @@ impl XPathCheckFlags {
     }
 }
 
-impl From<&Flags> for XPathCheckFlags {
-    fn from(flags: &Flags) -> Self {
+impl From<(Flags, Options)> for XPathCheckFlags {
+    fn from((flags, options): (Flags, Options)) -> Self {
         let mut check = Self::empty();
         if !flags.allow_unsafe_filename() {
             check.insert(Self::SAFE_NAME);
         }
-        if !flags.allow_unsafe_mkbdev() {
+        if !options.allow_unsafe_mkbdev() {
             check.insert(Self::RESTRICT_MKBDEV);
         }
         if flags.force_no_magiclinks() {
@@ -456,12 +456,6 @@ impl From<&Flags> for XPathCheckFlags {
     }
 }
 
-impl From<Flags> for XPathCheckFlags {
-    fn from(flags: Flags) -> Self {
-        (&flags).into()
-    }
-}
-
 impl XPath {
     /// Detects unsafe paths.
     ///
@@ -1036,6 +1030,9 @@ impl XPath {
             // 3. Confining memory file descriptors with
             //    `allow/create+!memfd:*` or
             //    `allow/truncate+!memfd-hugetlb:*`.
+            // 4. Confining secret memory file descriptors with
+            //    `allow/create+!secretmem` or
+            //    `allow/truncate+!secretmem`.
             Some(b'!') => self.is_special(),
             _ => {
                 // We match on canonicalized paths,
@@ -1053,6 +1050,7 @@ impl XPath {
         self.is_equal(b"!unnamed")
             || self.starts_with(b"!memfd:")
             || self.starts_with(b"!memfd-hugetlb:")
+            || self.is_equal(b"!secretmem")
     }
 
     /// Check if path starts with the `MAGIC_PREFIX`.
@@ -1123,6 +1121,18 @@ impl XPath {
         self.starts_with(b"/proc/")
     }
 
+    /// Check if path points to a KCOV path.
+    pub fn is_kcov(&self) -> bool {
+        cfg!(feature = "kcov") && self.is_equal(b"/dev/kcov")
+            || self.is_equal(b"/sys/kernel/debug/kcov")
+    }
+
+    /// Check if path points to a KCOV memfd path.
+    #[cfg(feature = "kcov")]
+    pub fn is_kcov_mfd(&self) -> bool {
+        self.is_equal(b"!memfd:syd-kcov")
+    }
+
     /// Check if path points to a static path.
     /// See proc_init in config.rs
     pub fn is_static(&self) -> bool {
