@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/proc.rs: /proc utilities
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 // Based in part upon procps-ng's library/devname.c which is:
 //   Copyright © 2011-2023 Jim Warner <james.warner@comcast.net>
 //   Copyright © 2017-2023 Craig Small <csmall@dropbear.xyz>
@@ -25,10 +25,11 @@
 #![deny(unsafe_code)]
 
 use std::{
+    borrow::Cow,
     ffi::{CStr, OsStr, OsString},
     fmt,
     fs::File,
-    io::{BufReader, Cursor, IoSlice, IoSliceMut, Read, Seek, SeekFrom},
+    io::{BufReader, Cursor, IoSlice, IoSliceMut, Read, Seek, SeekFrom, Write},
     num::NonZeroUsize,
     ops::Range,
     os::{
@@ -40,7 +41,7 @@ use std::{
 
 use bitflags::bitflags;
 use btoi::{btoi, btoi_radix};
-use memchr::{arch::all::is_prefix, memrchr};
+use memchr::{arch::all::is_prefix, memchr, memrchr};
 use nix::{
     errno::Errno,
     fcntl::OFlag,
@@ -50,7 +51,7 @@ use nix::{
         sysinfo::sysinfo,
         uio::{process_vm_readv, process_vm_writev, RemoteIoVec},
     },
-    unistd::{getpid, lseek64, write, Gid, Pid, Uid, Whence},
+    unistd::{getpid, gettid, lseek64, write, Gid, Pid, Uid, Whence},
     NixPath,
 };
 use procfs_core::{
@@ -63,21 +64,22 @@ use procfs_core::{
 use serde::{ser::SerializeMap, Serialize, Serializer};
 
 use crate::{
-    compat::{getdents64, openat2, statx, ResolveFlag, STATX_INO},
+    compat::{getdents64, openat2, pidfd_get_tgid, statx, OpenHow, ResolveFlag, STATX_INO},
     config::*,
     confine::SydMemoryMap,
-    cookie::{PROCMAP_QUERY_COOKIE_ARG3, PROCMAP_QUERY_COOKIE_ARG4, PROCMAP_QUERY_COOKIE_ARG5},
+    cookie::{CookieIdx, SYSCOOKIE_POOL},
     elf::ElfType,
     err::{err2no, proc_error_to_errno},
-    fs::{is_empty_file, is_open_fd, parse_fd, readlinkat, AT_BADFD},
+    fd::{is_empty_file, is_open_fd, is_proc, parse_fd, AT_BADFD},
+    fs::readlinkat,
     hash::{SydHashMap, SydHashSet, SydIndexSet},
     io::{read_all, read_buf, write_all},
-    lookup::{safe_open, safe_open_how_abs},
+    lookup::safe_open,
     parsers::{
         map_result,
         proc::{
-            parse_max_open_files, parse_stat, parse_statm, parse_status, parse_status_tgid,
-            parse_status_umask,
+            parse_max_open_files, parse_pidfd_info_pid, parse_stat, parse_statm, parse_status,
+            parse_status_tgid, parse_status_umask,
         },
         read_to_end,
     },
@@ -129,6 +131,41 @@ bitflags! {
     }
 }
 
+impl fmt::Display for ProcmapQueryFlags {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        if self.contains(Self::VMA_READABLE) {
+            f.write_str("r")?;
+        } else {
+            f.write_str("-")?;
+        }
+        if self.contains(Self::VMA_WRITABLE) {
+            f.write_str("w")?;
+        } else {
+            f.write_str("-")?;
+        }
+        if self.contains(Self::VMA_EXECUTABLE) {
+            f.write_str("x")?;
+        } else {
+            f.write_str("-")?;
+        }
+        if self.contains(Self::VMA_SHARED) {
+            f.write_str("s")?;
+        } else {
+            f.write_str("p")?;
+        }
+        Ok(())
+    }
+}
+
+impl Serialize for ProcmapQueryFlags {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        serializer.collect_str(self)
+    }
+}
+
 impl From<MMPermissions> for ProcmapQueryFlags {
     fn from(perms: MMPermissions) -> Self {
         let mut flags = Self::empty();
@@ -210,15 +247,15 @@ pub struct ProcmapQuery {
     /// (out)
     pub vma_offset: u64,
 
-    /// Backing file’s inode number, or zero if no backing file.
+    /// Backing file's inode number, or zero if no backing file.
     /// (out)
     pub inode: u64,
 
-    /// Backing file’s device major number, or zero if no backing file.
+    /// Backing file's device major number, or zero if no backing file.
     /// (out)
     pub dev_major: u32,
 
-    /// Backing file’s device minor number, or zero if no backing file.
+    /// Backing file's device minor number, or zero if no backing file.
     /// (out)
     pub dev_minor: u32,
 
@@ -317,9 +354,9 @@ pub fn procmap_query<Fd: AsFd>(
             fd.as_fd().as_raw_fd(),
             PROCMAP_QUERY,
             &mut q,
-            *PROCMAP_QUERY_COOKIE_ARG3,
-            *PROCMAP_QUERY_COOKIE_ARG4,
-            *PROCMAP_QUERY_COOKIE_ARG5,
+            SYSCOOKIE_POOL.get(CookieIdx::ProcmapQueryArg3),
+            SYSCOOKIE_POOL.get(CookieIdx::ProcmapQueryArg4),
+            SYSCOOKIE_POOL.get(CookieIdx::ProcmapQueryArg5),
         )
     })
     .map(|_| q)
@@ -519,7 +556,6 @@ pub fn proc_statm(pid: Pid) -> Result<Statm, Errno> {
 }
 
 /// Return the status of the given process.
-#[inline]
 pub fn proc_status(pid: Pid) -> Result<Status, Errno> {
     proc_status_read(proc_status_open(pid)?)
 }
@@ -571,6 +607,35 @@ pub fn proc_umask(pid: Pid) -> Result<Mode, Errno> {
     map_result(parse_status_umask(&data[..nread]))
 }
 
+/// Retrieve the TGID referred to by a pidfd.
+///
+/// Uses `PIDFD_GET_INFO` ioctl(2) on Linux-6.12 and newer,
+/// falling back to `/proc/self/fdinfo` parsing on older kernels.
+pub fn proc_pidfd_get_tgid<Fd: AsFd>(pidfd: Fd) -> Result<Pid, Errno> {
+    if *HAVE_PIDFD_GET_INFO {
+        return pidfd_get_tgid(&pidfd);
+    }
+
+    //
+    // Fallback: Read /proc/thread-self/fdinfo/<pidfd>, and parse the Pid: field.
+    //
+
+    // Prepare path to proc_pid_fdinfo(5).
+    let mut path = XPathBuf::from_pid(gettid())?;
+    path.try_reserve(b"/fdinfo".len()).or(Err(Errno::ENOMEM))?;
+    path.push(b"fdinfo");
+    path.push_fd(pidfd.as_fd().as_raw_fd());
+
+    // Open path without following symlinks.
+    let file = safe_open(PROC_FILE(), &path, OFLAG_PROC, RFLAG_PROC).map_err(proc_errno)?;
+
+    // A typical pidfd fdinfo is about 100–150 bytes.
+    let mut buf = [0u8; 256];
+    let nread = read_buf(file, &mut buf)?;
+
+    map_result(parse_pidfd_info_pid(&buf[..nread]))
+}
+
 /// Locate stack and return the stack memory address range.
 pub fn proc_stack(pid: Pid) -> Result<Range<u64>, Errno> {
     let maps = proc_maps(pid)?;
@@ -831,24 +896,7 @@ pub fn proc_mem(pid: Pid) -> Result<u64, Errno> {
     Ok(sum)
 }
 
-/// Returns the memory map of an address by reading _proc_pid_maps_(5).
-///
-/// Returns EFAULT if the address does not fall within any range.
-pub fn proc_mmap(pid: Pid, addr: u64) -> Result<SydMemoryMap, Errno> {
-    let maps = proc_maps(pid)?;
-
-    for mmap in maps {
-        // Check if the address falls within range.
-        if (mmap.0.address.0..mmap.0.address.1).contains(&addr) {
-            return Ok(mmap);
-        }
-    }
-
-    Err(Errno::EFAULT)
-}
-
 /// Return the system pipe max limit.
-#[inline]
 pub fn proc_pipemax() -> Result<libc::c_int, Errno> {
     proc_pipemax_read(proc_pipemax_open()?)
 }
@@ -869,10 +917,19 @@ pub fn proc_pipemax_read<Fd: AsFd>(fd: Fd) -> Result<libc::c_int, Errno> {
 
 /// Retrieves the value of /proc/sys/fs/file-max.
 pub fn proc_fs_file_max() -> Result<u64, Errno> {
-    // SAFETY: Informational command,
-    // runs before syd::config::proc_init.
-    #[expect(clippy::disallowed_methods)]
-    let fd = nix::fcntl::openat(AT_BADFD, "/proc/sys/fs/file-max", OFLAG_PROC, Mode::empty())?;
+    let fd = proc_open().and_then(|fd| {
+        safe_open(
+            fd,
+            c"sys/fs/file/max",
+            OFlag::O_RDONLY | OFlag::O_NOCTTY,
+            ResolveFlag::RESOLVE_NO_XDEV,
+        )
+    })?;
+
+    // Check the file is an empty file or bail out.
+    if !is_empty_file(&fd).unwrap_or(false) {
+        return Err(Errno::EBADFD);
+    }
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
@@ -883,10 +940,19 @@ pub fn proc_fs_file_max() -> Result<u64, Errno> {
 
 /// Retrieves the value of /proc/sys/fs/nr_open.
 pub fn proc_fs_nr_open() -> Result<u64, Errno> {
-    // SAFETY: Informational command,
-    // runs before syd::config::proc_init.
-    #[expect(clippy::disallowed_methods)]
-    let fd = nix::fcntl::openat(AT_BADFD, "/proc/sys/fs/nr_open", OFLAG_PROC, Mode::empty())?;
+    let fd = proc_open().and_then(|fd| {
+        safe_open(
+            fd,
+            c"sys/fs/nr_open",
+            OFlag::O_RDONLY | OFlag::O_NOCTTY,
+            ResolveFlag::RESOLVE_NO_XDEV,
+        )
+    })?;
+
+    // Check the file is an empty file or bail out.
+    if !is_empty_file(&fd).unwrap_or(false) {
+        return Err(Errno::EBADFD);
+    }
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
@@ -897,15 +963,19 @@ pub fn proc_fs_nr_open() -> Result<u64, Errno> {
 
 /// Retrieves the value of /proc/sys/vm/mmap_min_addr.
 pub fn proc_mmap_min_addr() -> Result<u64, Errno> {
-    // SAFETY: Informational command,
-    // runs before syd::config::proc_init.
-    #[expect(clippy::disallowed_methods)]
-    let fd = nix::fcntl::openat(
-        AT_BADFD,
-        "/proc/sys/vm/mmap_min_addr",
-        OFLAG_PROC,
-        Mode::empty(),
-    )?;
+    let fd = proc_open().and_then(|fd| {
+        safe_open(
+            fd,
+            c"sys/vm/mmap_min_addr",
+            OFlag::O_RDONLY | OFlag::O_NOCTTY,
+            ResolveFlag::RESOLVE_NO_XDEV,
+        )
+    })?;
+
+    // Check the file is an empty file or bail out.
+    if !is_empty_file(&fd).unwrap_or(false) {
+        return Err(Errno::EBADFD);
+    }
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
@@ -914,33 +984,21 @@ pub fn proc_mmap_min_addr() -> Result<u64, Errno> {
     btoi::<u64>(data[..nread].trim_ascii()).or(Err(Errno::EINVAL))
 }
 
-/// Sets the value of /proc/sys/kernel/pid_max.
-pub fn proc_pid_max(max: u64) -> Result<(), Errno> {
-    // SAFETY: System configuration command,
-    // runs during syd::config::proc_init.
-    #[expect(clippy::disallowed_methods)]
-    let fd = nix::fcntl::openat(
-        AT_BADFD,
-        "/proc/sys/kernel/pid_max",
-        OFLAG_PROC | OFlag::O_WRONLY,
-        Mode::empty(),
-    )?;
-
-    let mut buf = itoa::Buffer::new();
-    write_all(fd, buf.format(max).as_bytes())
-}
-
 /// Retrieves the value of /proc/sys/kernel/randomize_va_space
 pub fn proc_kernel_randomize_va_space() -> Result<u8, Errno> {
-    // SAFETY: Informational command,
-    // runs before syd::config::proc_init.
-    #[expect(clippy::disallowed_methods)]
-    let fd = nix::fcntl::openat(
-        AT_BADFD,
-        "/proc/sys/kernel/randomize_va_space",
-        OFLAG_PROC,
-        Mode::empty(),
-    )?;
+    let fd = proc_open().and_then(|fd| {
+        safe_open(
+            fd,
+            c"sys/kernel/randomize_va_space",
+            OFlag::O_RDONLY | OFlag::O_NOCTTY,
+            ResolveFlag::RESOLVE_NO_XDEV,
+        )
+    })?;
+
+    // Check the file is an empty file or bail out.
+    if !is_empty_file(&fd).unwrap_or(false) {
+        return Err(Errno::EBADFD);
+    }
 
     // Read up to 1 byte + 1 byte for \n.
     let mut data = [0u8; 2];
@@ -956,12 +1014,12 @@ pub fn proc_kernel_randomize_va_space() -> Result<u8, Errno> {
 pub fn proc_tty(pid: Pid) -> Result<XPathBuf, Errno> {
     let stat = proc_stat(pid)?;
     if stat.tty_nr <= 0 {
-        // Process has no controlling terminal
+        // Process has no controlling terminal.
         return Err(Errno::ENXIO);
     }
 
-    // minor is bits 31-20 and 7-0
-    // major is 15-8
+    // minor is bits 31-20 and 7-0.
+    // major is 15-8.
     let tty_nr = stat.tty_nr;
 
     // mmmmmmmmmmmm____MMMMMMMMmmmmmmmm
@@ -1261,14 +1319,32 @@ pub struct Vma {
 
     // VMA flags
     flags: ProcmapQueryFlags,
+
+    // VMA name
+    name: [u8; PATH_MAX],
 }
 
 impl Vma {
-    fn new(ptr: NonNull<libc::c_void>, len: NonZeroUsize, flags: ProcmapQueryFlags) -> Self {
-        Self { ptr, len, flags }
+    fn new(
+        ptr: NonNull<libc::c_void>,
+        len: NonZeroUsize,
+        flags: ProcmapQueryFlags,
+        name: [u8; PATH_MAX],
+    ) -> Self {
+        Self {
+            ptr,
+            len,
+            flags,
+            name,
+        }
     }
 
     /// Return VMA address.
+    pub fn addr(&self) -> usize {
+        self.ptr.as_ptr() as usize
+    }
+
+    /// Return VMA pointer.
     pub fn as_ptr(&self) -> NonNull<libc::c_void> {
         self.ptr
     }
@@ -1282,6 +1358,108 @@ impl Vma {
     pub fn flags(&self) -> ProcmapQueryFlags {
         self.flags
     }
+
+    /// Return VMA name as `XPath`.
+    pub fn name(&self) -> &XPath {
+        XPath::from_bytes(self.name_bytes())
+    }
+
+    /// Return VMA name as bytes.
+    pub fn name_bytes(&self) -> &[u8] {
+        let len = memchr(0, &self.name).unwrap_or(PATH_MAX);
+        &self.name[..len]
+    }
+}
+
+impl TryFrom<SydMemoryMap> for Vma {
+    type Error = Errno;
+
+    fn try_from(map: SydMemoryMap) -> Result<Self, Self::Error> {
+        let map = map.0;
+        let (start, end) = (map.address.0, map.address.1);
+        let ptr = NonNull::new(start as *mut _).ok_or(Errno::EINVAL)?;
+        let len = end
+            .checked_sub(start)
+            .ok_or(Errno::EINVAL)
+            .map(usize::try_from)?
+            .or(Err(Errno::EINVAL))
+            .map(NonZeroUsize::new)?
+            .ok_or(Errno::EINVAL)?;
+
+        let mut name = [0u8; PATH_MAX];
+        match &map.pathname {
+            MMapPath::Path(path) => {
+                let bytes = path.as_os_str().as_bytes();
+                let len = bytes.len().min(PATH_MAX);
+                name[..len].copy_from_slice(&bytes[..len]);
+            }
+            MMapPath::Heap => {
+                name[..7].copy_from_slice(b"[heap]\0");
+            }
+            MMapPath::Stack => {
+                name[..8].copy_from_slice(b"[stack]\0");
+            }
+            MMapPath::TStack(tid) => {
+                use std::io::Write;
+                let _ = write!(&mut name[..], "[stack:{tid}]\0");
+            }
+            MMapPath::Vdso => {
+                name[..7].copy_from_slice(b"[vdso]\0");
+            }
+            MMapPath::Vvar => {
+                name[..7].copy_from_slice(b"[vvar]\0");
+            }
+            MMapPath::Vsyscall => {
+                name[..11].copy_from_slice(b"[vsyscall]\0");
+            }
+            MMapPath::Rollup | MMapPath::Vsys(_) | MMapPath::Other(_) | MMapPath::Anonymous => {}
+        }
+
+        Ok(Self::new(ptr, len, map.perms.into(), name))
+    }
+}
+
+impl TryFrom<(ProcmapQuery, [u8; PATH_MAX])> for Vma {
+    type Error = Errno;
+
+    fn try_from((q, name): (ProcmapQuery, [u8; PATH_MAX])) -> Result<Self, Self::Error> {
+        let ptr = NonNull::new(q.vma_start as *mut _).ok_or(Errno::EINVAL)?;
+        let len = q
+            .vma_end
+            .checked_sub(q.vma_start)
+            .ok_or(Errno::EINVAL)
+            .map(usize::try_from)?
+            .or(Err(Errno::EINVAL))
+            .map(NonZeroUsize::new)?
+            .ok_or(Errno::EINVAL)?;
+        let flags = ProcmapQueryFlags::from_bits_truncate(q.vma_flags);
+        Ok(Self::new(ptr, len, flags, name))
+    }
+}
+
+impl fmt::Display for Vma {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let start = self.addr();
+        let end = start.saturating_add(self.len().get());
+        let flags = self.flags();
+        let name = self.name();
+        write!(f, "{start:x}-{end:x} {flags} {name}")
+    }
+}
+
+impl Serialize for Vma {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let start = self.addr();
+        let end = start.saturating_add(self.len().get());
+        let mut map = serializer.serialize_map(Some(3))?;
+        map.serialize_entry("addr", &[start, end])?;
+        map.serialize_entry("perm", &self.flags())?;
+        map.serialize_entry("name", &self.name())?;
+        map.end()
+    }
 }
 
 /// Get the VMA covering `addr` in process `pid`.
@@ -1289,7 +1467,7 @@ impl Vma {
 /// Tries the ioctl(2)-based fast path on Linux>=6.11,
 /// otherwise falls back to parsing _proc_pid_maps_(5).
 ///
-/// On success returns `(vma_start, vma_length, vma_flags)`.
+/// On success returns `Vma`.
 /// If no VMA covers `addr`, returns `Err(Errno::ENOENT)`.
 pub fn proc_get_vma(pid: Pid, addr: u64) -> Result<Vma, Errno> {
     if *HAVE_PROCMAP_QUERY {
@@ -1300,18 +1478,10 @@ pub fn proc_get_vma(pid: Pid, addr: u64) -> Result<Vma, Errno> {
     // Fallback: Sequentially scan /proc/$pid/maps.
     let maps = proc_maps(pid)?;
 
-    for SydMemoryMap(map) in maps {
-        let (start, end) = (map.address.0, map.address.1);
+    for map in maps {
+        let (start, end) = (map.0.address.0, map.0.address.1);
         if (start..end).contains(&addr) {
-            let ptr = NonNull::new(start as *mut _).ok_or(Errno::EINVAL)?;
-            let len = end
-                .checked_sub(start)
-                .ok_or(Errno::EINVAL)
-                .map(usize::try_from)?
-                .or(Err(Errno::EINVAL))
-                .map(NonZeroUsize::new)?
-                .ok_or(Errno::EINVAL)?;
-            return Ok(Vma::new(ptr, len, map.perms.into()));
+            return Vma::try_from(map);
         }
     }
 
@@ -1328,20 +1498,93 @@ fn procmap_query_get_vma(pid: Pid, addr: u64) -> Result<Vma, Errno> {
     // Call PROCMAP_QUERY ioctl(2).
     //
     // No name or build-id buffers needed.
-    let q = procmap_query(&maps, flags, addr, None, None)?;
+    let mut name = [0u8; PATH_MAX];
+    let q = procmap_query(&maps, flags, addr, Some(&mut name), None)?;
 
-    let ptr = NonNull::new(q.vma_start as *mut _).ok_or(Errno::EINVAL)?;
-    let len = q
-        .vma_end
-        .checked_sub(q.vma_start)
-        .ok_or(Errno::EINVAL)
-        .map(usize::try_from)?
-        .or(Err(Errno::EINVAL))
-        .map(NonZeroUsize::new)?
-        .ok_or(Errno::EINVAL)?;
-    let flags = ProcmapQueryFlags::from_bits_truncate(q.vma_flags);
+    Vma::try_from((q, name))
+}
 
-    Ok(Vma::new(ptr, len, flags))
+/// Get all VMAs matching `flags` in process `pid`.
+///
+/// Tries the ioctl(2)-based fast path on Linux>=6.11,
+/// otherwise falls back to parsing _proc_pid_maps_(5).
+///
+/// On success returns `Vec<Vma>`.
+pub fn proc_find_vma(pid: Pid, flags: ProcmapQueryFlags) -> Result<Vec<Vma>, Errno> {
+    if *HAVE_PROCMAP_QUERY {
+        // Use the efficient PROCMAP_QUERY ioctl(2) on Linux>=6.11.
+        return procmap_query_find_vma(pid, flags);
+    }
+
+    // Fallback: Sequentially scan /proc/$pid/maps.
+    let maps = proc_maps(pid)?;
+    let mut vmas = Vec::new();
+
+    for map in maps {
+        let perms = ProcmapQueryFlags::from(map.0.perms);
+
+        if flags.contains(ProcmapQueryFlags::VMA_READABLE)
+            && !perms.contains(ProcmapQueryFlags::VMA_READABLE)
+        {
+            continue;
+        }
+
+        if flags.contains(ProcmapQueryFlags::VMA_WRITABLE)
+            && !perms.contains(ProcmapQueryFlags::VMA_WRITABLE)
+        {
+            continue;
+        }
+
+        if flags.contains(ProcmapQueryFlags::VMA_EXECUTABLE)
+            && !perms.contains(ProcmapQueryFlags::VMA_EXECUTABLE)
+        {
+            continue;
+        }
+
+        if flags.contains(ProcmapQueryFlags::VMA_SHARED)
+            && !perms.contains(ProcmapQueryFlags::VMA_SHARED)
+        {
+            continue;
+        }
+
+        if flags.contains(ProcmapQueryFlags::FILE_BACKED_VMA)
+            && !matches!(map.0.pathname, MMapPath::Path(_))
+        {
+            continue;
+        }
+
+        vmas.push(Vma::try_from(map)?);
+    }
+
+    Ok(vmas)
+}
+
+/// Get all VMAs matching `flags` in process `pid` using PROCMAP_QUERY ioctl(2).
+pub fn procmap_query_find_vma(pid: Pid, flags: ProcmapQueryFlags) -> Result<Vec<Vma>, Errno> {
+    // Open /proc/$pid/maps.
+    let maps = proc_maps_open(pid)?;
+
+    let query_flags = flags | ProcmapQueryFlags::COVERING_OR_NEXT_VMA;
+
+    let mut vmas = Vec::new();
+    let mut addr = 0u64;
+
+    loop {
+        // Call PROCMAP_QUERY ioctl(2).
+        //
+        // No name or build-id buffers needed.
+        let mut name = [0u8; PATH_MAX];
+        match procmap_query(&maps, query_flags, addr, Some(&mut name), None) {
+            Ok(q) => {
+                vmas.push(Vma::try_from((q, name))?);
+                addr = q.vma_end;
+            }
+            Err(Errno::ENOENT) => break,
+            Err(errno) => return Err(errno),
+        }
+    }
+
+    Ok(vmas)
 }
 
 /// Parses /proc/pid/environ and returns a HashMap.
@@ -1503,11 +1746,12 @@ pub fn proc_unix_inodes(pid: Pid) -> Result<SydHashSet<u64>, Errno> {
 ///
 /// * `pid` - The PID of the target process.
 /// * `elf_type` - The ELF type (Elf32 or Elf64) of the target process.
+/// * `deny_vdso` - True to deny access to vDSO by overwriting AT_SYSINFO{,_EHDR}.
 ///
 /// # Returns
 ///
 /// Returns `Ok(())` on success, or an `Err(Errno)` if an error occurs.
-pub fn proc_set_at_secure(pid: Pid, elf_type: ElfType) -> Result<(), Errno> {
+pub fn proc_set_at_secure(pid: Pid, elf_type: ElfType, deny_vdso: bool) -> Result<(), Errno> {
     // This function requires CONFIG_CROSS_MEMORY_ATTACH,
     // unless Syd is explicitly configured to fallback.
     let mut proc_mem = if Sandbox::use_proc_pid_mem() {
@@ -1678,6 +1922,14 @@ pub fn proc_set_at_secure(pid: Pid, elf_type: ElfType) -> Result<(), Errno> {
     const AT_REQKEY: &[usize] = &[AT_UID, AT_EUID, AT_GID, AT_EGID, AT_SECURE];
     let mut required_index = 0;
 
+    // Used to deny access to vDSO as necessary.
+    //
+    // 1. libc crate does not define these for all arches we support,
+    //    however the values are consistent across all.
+    // 2. AT_SECURE comes after AT_SYSINFO{,_EHDR} in the auxiliary vector.
+    const AT_SYSINFO: usize = 32;
+    const AT_SYSINFO_EHDR: usize = 33;
+
     // Store the UID/GID values for comparison.
     // We're going to check these values against ours
     // and fail as necessary.
@@ -1765,6 +2017,28 @@ pub fn proc_set_at_secure(pid: Pid, elf_type: ElfType) -> Result<(), Errno> {
         } else if required_index > 0 {
             // An unexpected key appeared; validation fails!
             return Err(Errno::EACCES);
+        } else if deny_vdso && (key == AT_SYSINFO || key == AT_SYSINFO_EHDR) {
+            // Deny access to vDSO as requested.
+            // AT_SYS_INFO{,_EHDR} always comes before AT_SECURE.
+            //
+            // Overwrite the key and value in the local buffer:
+            // Overwrite the key with AT_SECURE.
+            // Overwrite the value with 1.
+            let key = usize_to_ne_bytes(AT_SECURE, sizeof_ptr);
+            let val = usize_to_ne_bytes(1, sizeof_ptr);
+            buf[offset..key_end].copy_from_slice(&key);
+            buf[key_end..val_end].copy_from_slice(&val);
+
+            // Write the modified value back to the target process.
+            if proc_write(
+                proc_mem.as_mut(),
+                pid,
+                &buf[offset..val_end],
+                sp.checked_add(offset).ok_or(Errno::EINVAL)?,
+            )? != sizeof_ptr2
+            {
+                return Err(Errno::EIO);
+            }
         } else if key == AT_NULL {
             break;
         }
@@ -1841,7 +2115,7 @@ pub fn proc_max_open_files(pid: Pid) -> Result<LimitValue, Errno> {
     map_result(parse_max_open_files(read_to_end(file, &mut buf)?))
 }
 
-/// Pick a random fd number and check via KCMP_FILE whether it’s open in `pid`.
+/// Pick a random fd number and check via KCMP_FILE whether it's open in `pid`.
 /// On success returns the fd; if none are found after probing the full range,
 /// returns `Err(Errno::EBADF)`. Other errors (permissions, invalid pid, etc.)
 /// are returned immediately.
@@ -1996,15 +2270,19 @@ impl fmt::Display for KernelTaintFlags {
 
 /// Retrieves the value of /proc/sys/kernel/tainted as KernelTaintFlags.
 pub fn proc_kernel_taint() -> Result<KernelTaintFlags, Errno> {
-    // SAFETY: Informational command,
-    // runs before syd::config::proc_init.
-    #[expect(clippy::disallowed_methods)]
-    let fd = nix::fcntl::openat(
-        AT_BADFD,
-        "/proc/sys/kernel/tainted",
-        OFLAG_PROC,
-        Mode::empty(),
-    )?;
+    let fd = proc_open().and_then(|fd| {
+        safe_open(
+            fd,
+            c"sys/kernel/tainted",
+            OFlag::O_RDONLY | OFlag::O_NOCTTY,
+            ResolveFlag::RESOLVE_NO_XDEV,
+        )
+    })?;
+
+    // Check the file is an empty file or bail out.
+    if !is_empty_file(&fd).unwrap_or(false) {
+        return Err(Errno::EBADFD);
+    }
 
     // Read up to 24 bytes + 1 byte for \n.
     let mut data = [0u8; 25];
@@ -2015,17 +2293,90 @@ pub fn proc_kernel_taint() -> Result<KernelTaintFlags, Errno> {
     Ok(KernelTaintFlags::from_bits_retain(val))
 }
 
-/// Write the given $data to /proc/$pid/$name.
-pub fn proc_write(pid: Pid, name: &[u8], data: &[u8]) -> Result<(), Errno> {
-    let mut pfd = XPathBuf::from("/proc");
-    pfd.push_pid(pid);
-    pfd.push(name);
+/// Set max PID limit for PID namespace.
+pub fn proc_pid_max<Fd: AsFd>(fd_proc: Fd, pid_max: u64) -> Result<(), Errno> {
+    let mut buf = itoa::Buffer::new();
+    let pid_max = buf.format(pid_max).as_bytes();
+    proc_write(&fd_proc, b"sys/kernel/pid/max", pid_max, false /*pid*/)
+}
 
-    let how = safe_open_how_abs(OFlag::O_WRONLY, ResolveFlag::empty());
+/// Set up user mappings for user namespace.
+pub fn proc_map_user<Fd: AsFd>(
+    fd_proc: Fd,
+    uid: Uid,
+    gid: Gid,
+    map_root: bool,
+) -> Result<(), Errno> {
+    let uid_map = {
+        let mut buf = Vec::new();
+        write!(
+            &mut buf,
+            "{} {uid} 1",
+            if map_root { 0 } else { uid.as_raw() }
+        )
+        .map_err(|err| err2no(&err))?;
+        buf
+    };
+    let gid_map = {
+        let mut buf = Vec::new();
+        write!(
+            &mut buf,
+            "{} {gid} 1",
+            if map_root { 0 } else { gid.as_raw() }
+        )
+        .map_err(|err| err2no(&err))?;
+        buf
+    };
 
-    // This function runs early at startup before syd::config::proc_init.
-    #[expect(clippy::disallowed_methods)]
-    let fd = retry_on_eintr(|| openat2(AT_BADFD, &pfd, how))?;
+    // Write "deny" to /proc/self/setgroups before writing to gid_map.
+    proc_write(&fd_proc, b"setgroups", b"deny", true /*pid*/)?;
+
+    // Write GID mapping to gid_map.
+    proc_write(&fd_proc, b"gid_map", &gid_map, true /*pid*/)?;
+
+    // Write UID mapping to uid map.
+    proc_write(&fd_proc, b"uid_map", &uid_map, true /*pid*/)?;
+
+    Ok(())
+}
+
+/// Set up time offsets for time namespace.
+pub fn proc_set_time<Fd: AsFd>(fd_proc: Fd, boffset: i64, moffset: i64) -> Result<(), Errno> {
+    // <clock-id> <offset-secs> <offset-nanosecs>
+    let mut buf = itoa::Buffer::new();
+    let mut data = "monotonic ".to_string();
+    data.push_str(buf.format(moffset));
+    data.push_str(" 0\nboottime ");
+    data.push_str(buf.format(boffset));
+    data.push_str(" 0\n");
+
+    proc_write(
+        fd_proc,
+        b"timens_offsets",
+        data.as_bytes(),
+        true, /*pid*/
+    )
+}
+
+// Write the given $data to /proc/$pid/$name.
+fn proc_write<Fd: AsFd>(fd_proc: Fd, name: &[u8], data: &[u8], pid: bool) -> Result<(), Errno> {
+    let pfd = if pid {
+        let mut pfd = XPathBuf::from_pid(Pid::this())?;
+        pfd.try_reserve(name.len().checked_add(1).ok_or(Errno::EOVERFLOW)?)
+            .or(Err(Errno::ENOMEM))?;
+        pfd.push(name);
+        Cow::Owned(pfd)
+    } else {
+        Cow::Borrowed(XPath::from_bytes(name))
+    };
+
+    // Open proc file safely.
+    let fd = safe_open(
+        fd_proc,
+        pfd.as_ref(),
+        OFlag::O_WRONLY,
+        ResolveFlag::RESOLVE_NO_XDEV,
+    )?;
 
     // Check the file is an empty file or bail out.
     if !is_empty_file(&fd).unwrap_or(false) {
@@ -2035,6 +2386,39 @@ pub fn proc_write(pid: Pid, name: &[u8], data: &[u8]) -> Result<(), Errno> {
     write_all(&fd, data)
 }
 
+/// Open /proc safely and validate procfs.
+pub fn proc_open() -> Result<OwnedFd, Errno> {
+    let how = OpenHow::new()
+        .flags(OFlag::O_PATH | OFlag::O_DIRECTORY | OFlag::O_NOFOLLOW | OFlag::O_CLOEXEC)
+        .resolve(ResolveFlag::RESOLVE_NO_MAGICLINKS | ResolveFlag::RESOLVE_NO_SYMLINKS);
+    // This function is called before syd::config::proc_init.
+    #[expect(clippy::disallowed_methods)]
+    let fd_proc = openat2(AT_BADFD, "/proc", how)?;
+
+    // SAFETY: Validate what we've opened is procfs(5).
+    if !is_proc(&fd_proc).unwrap_or(false) {
+        return Err(Errno::ENODEV);
+    }
+
+    Ok(fd_proc)
+}
+
+#[cfg(feature = "kcov")]
+pub(crate) fn proc_kcov_read_id(tid: Pid, fd: RawFd) -> Result<crate::kcov::KcovId, Errno> {
+    if tid.as_raw() < 1 {
+        return Err(Errno::EINVAL);
+    }
+    if fd < 0 {
+        return Err(Errno::EBADF);
+    }
+
+    let pfd = XPathBuf::from_pid_fd(tid, fd)?;
+
+    statx(PROC_FILE(), &pfd, 0, STATX_INO)
+        .map(|statx| statx.stx_ino)
+        .map(crate::kcov::KcovId::new)
+}
+
 #[cfg(test)]
 mod tests {
     use std::os::unix::fs::PermissionsExt;
@@ -2042,6 +2426,7 @@ mod tests {
     use nix::sys::stat::umask;
 
     use super::*;
+    use crate::fd::pidfd_open;
 
     fn setup() -> bool {
         let _ = crate::log::log_init_simple(crate::syslog::LogLevel::Warn);
@@ -2475,4 +2860,36 @@ mod tests {
     fn test_proc_kernel_randomize_va_space() {
         matches!(proc_kernel_randomize_va_space(), Ok(0 | 1 | 2));
     }
+
+    #[test]
+    fn test_proc_find_vma() {
+        if !setup() {
+            return;
+        }
+
+        let pid = Pid::this();
+        for flags in [
+            ProcmapQueryFlags::VMA_READABLE,
+            ProcmapQueryFlags::VMA_EXECUTABLE,
+            ProcmapQueryFlags::VMA_READABLE | ProcmapQueryFlags::VMA_EXECUTABLE,
+        ] {
+            let vmas = proc_find_vma(pid, flags).unwrap();
+            assert!(!vmas.is_empty());
+            for vma in vmas {
+                assert!(vma.flags().contains(flags));
+            }
+        }
+    }
+
+    #[test]
+    fn test_proc_pidfd_get_tgid() {
+        if !setup() {
+            return;
+        }
+
+        let mypid = getpid();
+        let pidfd = pidfd_open(mypid, 0).unwrap();
+        let retpid = proc_pidfd_get_tgid(pidfd).unwrap();
+        assert_eq!(mypid, retpid);
+    }
 }
