@@ -39,6 +39,7 @@ use nix::{
 use crate::{
     compat::{fstatx, MFdFlags, STATX_TYPE},
     config::ENV_SKIP_SCMP,
+    confine::secure_getenv,
     err::err2no,
     io::ReadFd,
     lookup::FileType,
@@ -153,7 +154,7 @@ impl SealedCommand {
             Some(val) => env::set_var("SYD_RUST_BACKTRACE", val),
             None => env::remove_var("SYD_RUST_BACKTRACE"),
         };
-        if env::var_os(ENV_SKIP_SCMP).is_none() {
+        if secure_getenv(ENV_SKIP_SCMP).is_none() {
             env::set_var("RUST_BACKTRACE", "0");
         }
 
