@@ -600,6 +600,11 @@ impl<T: Copy> Sealable<T> {
             Sealable::Sealed(s) => s,
         }
     }
+
+    /// Return true if sealed.
+    pub fn is_sealed(&self) -> bool {
+        matches!(self, Sealable::Sealed(_))
+    }
 }
 
 impl<T: Copy> Deref for Sealable<T> {
