@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/test/util.rs: Utilities for integration tests
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -158,6 +158,14 @@ macro_rules! assert_status_aborted {
     }};
 }
 
+#[macro_export]
+macro_rules! assert_status_terminated {
+    ($status:expr) => {{
+        const XTERM: i32 = 128 + libc::SIGTERM;
+        assert_status_code_matches!($status, XTERM);
+    }};
+}
+
 #[macro_export]
 macro_rules! assert_status_panicked {
     ($status:expr) => {{
@@ -239,6 +247,13 @@ macro_rules! assert_status_permission_denied {
     };
 }
 
+#[macro_export]
+macro_rules! assert_status_no_such_process {
+    ($status:expr) => {
+        $crate::assert_status_code!($status, libc::ESRCH);
+    };
+}
+
 #[macro_export]
 macro_rules! assert_status_broken_pipe {
     ($status:expr) => {
@@ -260,6 +275,13 @@ macro_rules! assert_status_connection_refused {
     };
 }
 
+#[macro_export]
+macro_rules! assert_status_faulted {
+    ($status:expr) => {
+        $crate::assert_status_code!($status, libc::EFAULT);
+    };
+}
+
 #[macro_export]
 macro_rules! assert_status_hidden {
     ($status:expr) => {
@@ -274,6 +296,20 @@ macro_rules! assert_status_notdir {
     };
 }
 
+#[macro_export]
+macro_rules! assert_status_notty {
+    ($status:expr) => {
+        $crate::assert_status_code!($status, libc::ENOTTY);
+    };
+}
+
+#[macro_export]
+macro_rules! assert_status_illegal_sequence {
+    ($status:expr) => {
+        $crate::assert_status_code!($status, libc::EILSEQ);
+    };
+}
+
 #[macro_export]
 macro_rules! assert_status_invalid {
     ($status:expr) => {
@@ -315,6 +351,19 @@ macro_rules! assert_status_killed {
     };
 }
 
+#[macro_export]
+macro_rules! assert_status_not_killed {
+    ($status:expr) => {
+        let code = $status.code().unwrap_or(127);
+        $crate::assert_ne!(
+            code,
+            128 + libc::SIGKILL,
+            "code:{code} status:{:?}",
+            $status
+        );
+    };
+}
+
 #[macro_export]
 macro_rules! assert_status_sigsys {
     ($status:expr) => {
@@ -391,6 +440,17 @@ macro_rules! skip_unless_available {
     }};
 }
 
+#[macro_export]
+macro_rules! skip_unless_trusted {
+    () => {{
+        if !cfg!(feature = "trusted") {
+            eprintln!("trusted feature not enabled, skipping!");
+            std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
+            return Ok(());
+        }
+    }};
+}
+
 #[macro_export]
 macro_rules! skip_if_32bin {
     () => {{
@@ -442,6 +502,24 @@ macro_rules! skip_if_root {
     }};
 }
 
+#[macro_export]
+macro_rules! skip_if_fs {
+    ($($fs:expr),* $(,)?) => {{
+        let cwd = open(".", OFlag::O_PATH | OFlag::O_DIRECTORY, Mode::empty()).unwrap();
+        let current_fs = syd::compat::FsType::get(cwd).unwrap();
+        $(
+            if let Ok(fs_types) = syd::compat::FsType::from_name($fs) {
+                if fs_types.contains(&current_fs) {
+                    eprintln!("Test does not run correctly on {}, skipping!", $fs);
+                    std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
+                    return Ok(());
+                }
+            }
+        )*
+        eprintln!("Filesystem check passed, continuing...");
+    }};
+}
+
 #[macro_export]
 macro_rules! skip_unless_iproute2 {
     () => {{
@@ -661,6 +739,14 @@ macro_rules! skip_unless_unshare {
             return Ok(());
         }
 
+        // Apparmor does not allow use of unprivileged namespaces by default.
+        if syd::confine::apparmor_enabled().unwrap_or(false) {
+            eprintln!("Apparmor detected!");
+            eprintln!("Test requires Linux namespaces ({:?}), skipping!", [$($ns),*]);
+            std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
+            return Ok(());
+        }
+
         match syd::confine::ns_enabled(flags) {
             Ok(false) => {
                 eprintln!("Test requires Linux namespaces ({:?}), skipping!", [$($ns),*]);
@@ -741,23 +827,29 @@ macro_rules! skip_unless_cap {
     }};
 }
 
+#[macro_export]
+macro_rules! skip_if_landlock_abi_supported {
+    ($abi:expr) => {{
+        if syd::landlock::ABI::new_current() >= syd::landlock::ABI::from($abi) {
+            eprintln!(
+                "Test is for Landlock ABI versions older than {}, skipping!",
+                $abi
+            );
+            std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
+            return Ok(());
+        } else {
+            eprintln!(
+                "Landlock ABI {} is NOT supported, proceeding with test...",
+                $abi
+            );
+        }
+    }};
+}
+
 #[macro_export]
 macro_rules! skip_unless_landlock_abi_supported {
     ($abi:expr) => {{
-        use std::process::Command;
-
-        // Run `syd-lock -A` to get the ABI version as exit code.
-        let status = Command::new(&*SYD_LOCK)
-            .arg("-A")
-            .status()
-            .expect("Failed to run syd-lock -A");
-
-        // SAFETY: We expect `syd-lock -A` to exit with a code.
-        let exit_code = status
-            .code()
-            .expect("Failed to get exit code from syd-lock -A");
-
-        if exit_code < $abi {
+        if syd::landlock::ABI::new_current() < syd::landlock::ABI::from($abi) {
             eprintln!("Test requires Landlock ABI version {}, skipping!", $abi);
             std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
             return Ok(());
@@ -867,6 +959,7 @@ pub static SYD_DNS: LazyLock<String> =
 
 /// Holds a `String` to run `syd-env`.
 /// Honours CARGO_BIN_EXE_syd-env environment variable.
+#[allow(dead_code)]
 pub static SYD_ENV: LazyLock<String> =
     LazyLock::new(|| env::var("CARGO_BIN_EXE_syd-env").unwrap_or("syd-env".to_string()));
 
@@ -981,6 +1074,7 @@ pub static CI_BUILD: LazyLock<bool> = LazyLock::new(|| env::var("JOB_ID").ok().i
 pub static GL_BUILD: LazyLock<bool> = LazyLock::new(|| env::var("CI_PROJECT_ID").ok().is_some());
 
 /// A boolean which specifies whether standard error is a TTY.
+#[allow(dead_code)]
 pub static SYD_TEST_TTY: LazyLock<bool> =
     LazyLock::new(|| isatty(std::io::stderr()).unwrap_or(false));
 
@@ -1096,6 +1190,7 @@ impl Syd {
         self
     }
 
+    #[allow(dead_code)]
     pub fn quiet(&mut self) -> &mut Self {
         self.stderr(Stdio::null());
         self.stdout(Stdio::null());
@@ -1158,9 +1253,11 @@ pub fn syd() -> Syd {
         LazyLock::new(|| env::var_os("SYD_TEST_HELGRIND").is_some());
     let mut cmd = Syd::new("timeout");
     if check_timeout_foreground() {
+        // Use -v, not --verbose.
+        // Latter is not supported by Android.
+        cmd.arg("-v");
         cmd.arg("--foreground");
         cmd.arg("--preserve-status");
-        cmd.arg("--verbose");
     }
     cmd.arg("-sKILL");
     cmd.arg(env::var("SYD_TEST_TIMEOUT").unwrap_or("15m".to_string()));
@@ -1628,14 +1725,20 @@ pub fn current_dir(base: bool) -> std::io::Result<PathBuf> {
 
 /// Shuffles a vector using the Fisher-Yates algorithm,
 /// utilizing random numbers obtained via `libc::rand`.
+/// On Android, falls back to `syd::rng::randint`.
 ///
 /// # Arguments
 /// * `vec` - A mutable reference to the vector to shuffle.
 pub fn shuffle_vec<T>(vec: &mut [T]) {
     let len = vec.len();
     for i in 0..len {
+        // Android does not define libc::rand().
+        #[cfg(not(target_os = "android"))]
         let r = unsafe { libc::rand() } as usize;
+        #[cfg(target_os = "android")]
+        let r = syd::rng::randint(usize::MIN..=usize::MAX).unwrap();
+
         let j = r % (len - i) + i;
-        vec.swap(i, j); // perform the swap
+        vec.swap(i, j); // Perform the swap.
     }
 }
