@@ -27,11 +27,12 @@ use crate::{
     caps::Capability,
     compat::{set_name, set_pdeathsig},
     config::*,
-    confine::{confine_scmp_ioctl_cld, confine_scmp_kptr, confine_scmp_pwritev2, safe_drop_cap},
-    err::err2set,
+    confine::{
+        confine_landlock_scope, confine_scmp_ioctl_cld, confine_scmp_kptr, confine_scmp_pwritev2,
+        safe_drop_cap,
+    },
     fd::close,
     ignore_signal,
-    landlock::{CompatLevel, Compatible, RestrictSelfFlags, Ruleset, RulesetAttr, Scope},
     retry::retry_on_eintr,
     unshare::{config::Config, error::ErrorCode as Err, run::ChildInfo},
 };
@@ -168,21 +169,8 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
     // 2. Signal send outside Landlock.
     // We leave path and network restrictions for Landlock
     // to be configured by the user using Lock sandboxing.
-    let compat_level = if *HAVE_LANDLOCK_SCOPED_SIGNALS {
-        CompatLevel::HardRequirement
-    } else {
-        CompatLevel::BestEffort
-    };
-    if let Err(error) = Ruleset::default()
-        .set_compatibility(compat_level)
-        .scope(Scope::Signal)
-        .and_then(|ruleset| ruleset.create())
-        .and_then(|ruleset| ruleset.restrict_self(RestrictSelfFlags::empty()))
-    {
-        if compat_level == CompatLevel::HardRequirement {
-            let errno = err2set(&error);
-            fail_errno_safe!(child, Err::LandlockFilterScopedSignals, errno as i32);
-        }
+    if let Err(errno) = confine_landlock_scope() {
+        fail_errno_safe!(child, Err::LandlockFilterScopedSignals, errno as i32);
     }
 
     // Restriction 2:
