@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/workers/aes.rs: `syd_aes' encryption thread
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -13,7 +13,7 @@
 
 use std::{
     collections::hash_map::Entry,
-    env, fmt,
+    fmt,
     os::fd::{AsFd, FromRawFd, OwnedFd, RawFd},
     sync::{
         atomic::{AtomicBool, Ordering},
@@ -26,38 +26,38 @@ use libseccomp::{scmp_cmp, RawSyscall, ScmpAction, ScmpFilterContext, ScmpSyscal
 use nix::{
     errno::Errno,
     fcntl::{posix_fadvise, splice, tee, OFlag, PosixFadviseAdvice, SpliceFFlags},
+    sched::{unshare, CloneFlags},
     unistd::{lseek64, write, Gid, Uid, Whence},
 };
 use serde::{Serialize, Serializer};
 
 #[cfg(target_arch = "x86")]
-use crate::cookie::FTRUNCATE64_COOKIE_ARG3;
+use crate::cookie::CookieIdx::Ftruncate64Arg3;
 use crate::{
+    alert,
     config::*,
     confine::{
-        confine_scmp_clone, confine_scmp_clone3, confine_scmp_fadvise, confine_scmp_madvise,
-        confine_scmp_write, scmp_add_setid_rules, ExportMode,
-    },
-    cookie::{
-        safe_ftruncate64, safe_pipe2, ACCEPT4_COOKIE_ARG4, ACCEPT4_COOKIE_ARG5,
-        FTRUNCATE64_COOKIE_ARG4, FTRUNCATE64_COOKIE_ARG5, FTRUNCATE_COOKIE_ARG2,
-        FTRUNCATE_COOKIE_ARG3, FTRUNCATE_COOKIE_ARG4, FTRUNCATE_COOKIE_ARG5, PIPE2_COOKIE_ARG2,
-        PIPE2_COOKIE_ARG3, PIPE2_COOKIE_ARG4, PIPE2_COOKIE_ARG5, SYS_ACCEPT4,
+        confine_scmp_clone, confine_scmp_clone3, confine_scmp_fadvise, confine_scmp_fcntl,
+        confine_scmp_madvise, confine_scmp_pipe2, confine_scmp_prctl, confine_scmp_sigaction,
+        confine_scmp_write, scmp_add_setid_rules, secure_getenv, ExportMode,
     },
+    cookie::{safe_ftruncate64, safe_pipe2, CookieIdx, SYSCOOKIE_POOL, SYS_ACCEPT4},
     debug,
     err::{err2no, SydJoinHandle, SydResult},
     error,
-    fs::seal_memfd_all,
+    fd::seal_memfd_all,
     hash::{
         aes_ctr_enc, aes_ctr_init, hmac_sha256_feed, hmac_sha256_fini, hmac_sha256_init,
         SydHashMap, BLOCK_SIZE, HMAC_TAG_SIZE, IV, IV_SIZE, SYD3_HDR_SIZE,
     },
     info,
+    landlock::Errata,
+    landlock_policy::LandlockPolicy,
     lookup::FileInfo,
     ofd::lock_fd,
     path::{XPath, XPathBuf},
     retry::{retry_on_eintr, retry_on_intr},
-    sandbox::Flags,
+    sandbox::Options,
 };
 
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Ord, PartialOrd)]
@@ -116,7 +116,7 @@ pub(crate) type AesLock = Arc<(Mutex<AesMap>, Condvar)>;
 pub(crate) struct AesWorker {
     fdalg: (RawFd, RawFd),
     files: AesLock,
-    flags: Flags,
+    options: Options,
     is_memfd: bool,
     should_exit: Arc<AtomicBool>,
     transit_uids: Vec<(Uid, Uid)>,
@@ -140,7 +140,7 @@ impl AesWorker {
     pub(crate) fn new(
         fdalg: (RawFd, RawFd),
         files: AesLock,
-        flags: Flags,
+        options: Options,
         is_memfd: bool,
         should_exit: Arc<AtomicBool>,
         transit_uids: &[(Uid, Uid)],
@@ -149,7 +149,7 @@ impl AesWorker {
         Self {
             fdalg,
             files,
-            flags,
+            options,
             is_memfd,
             should_exit,
             transit_uids: transit_uids.to_vec(),
@@ -160,10 +160,23 @@ impl AesWorker {
     /// Confine AES thread.
     #[expect(clippy::cognitive_complexity)]
     pub(crate) fn prepare_confine(
-        flags: Flags,
+        options: Options,
         transit_uids: &[(Uid, Uid)],
         transit_gids: &[(Gid, Gid)],
+        dry_run: bool,
     ) -> SydResult<ScmpFilterContext> {
+        if !dry_run {
+            // SAFETY: Set up a Landlock sandbox to disallow all access.
+            let abi = crate::landlock::ABI::new_current();
+            let errata = crate::landlock::Errata::query();
+            let policy = LandlockPolicy {
+                scoped_abs: true,
+                scoped_sig: errata.contains(Errata::SCOPED_SIGNAL_SAME_TGID),
+                ..Default::default()
+            };
+            let _ = policy.restrict_self(abi);
+        }
+
         // Create seccomp filter with default action.
         let mut ctx = ScmpFilterContext::new(ScmpAction::KillProcess)?;
 
@@ -173,7 +186,7 @@ impl AesWorker {
 
         // Disable Speculative Store Bypass mitigations
         // with trace/allow_unsafe_exec_speculative:1
-        ctx.set_ctl_ssb(flags.allow_unsafe_exec_speculative())?;
+        ctx.set_ctl_ssb(options.allow_unsafe_exec_speculative())?;
 
         // DO NOT synchronize filter to all threads.
         // Main thread will confine itself.
@@ -185,6 +198,10 @@ impl AesWorker {
         // Use a binary tree sorted by syscall number if possible.
         let _ = ctx.set_ctl_optimize(2);
 
+        // Syscall argument cookies may be disabled
+        // at startup with trace/allow_unsafe_nocookie:1.
+        let restrict_cookie = !options.allow_unsafe_nocookie();
+
         // SAFETY: Do NOT add supported architectures to the filter.
         // This ensures Syd can never run a non-native system call,
         // which we do not need at all.
@@ -269,12 +286,12 @@ impl AesWorker {
             Ok(syscall) => {
                 // Secure using syscall argument cookies.
                 let mut rules = vec![];
-                if !flags.allow_unsafe_nocookie() {
+                if restrict_cookie {
                     rules.extend(&[
-                        scmp_cmp!($arg2 == (*FTRUNCATE_COOKIE_ARG2).into()),
-                        scmp_cmp!($arg3 == (*FTRUNCATE_COOKIE_ARG3).into()),
-                        scmp_cmp!($arg4 == (*FTRUNCATE_COOKIE_ARG4).into()),
-                        scmp_cmp!($arg5 == (*FTRUNCATE_COOKIE_ARG5).into()),
+                        scmp_cmp!($arg2 == SYSCOOKIE_POOL.get(CookieIdx::FtruncateArg2).into()),
+                        scmp_cmp!($arg3 == SYSCOOKIE_POOL.get(CookieIdx::FtruncateArg3).into()),
+                        scmp_cmp!($arg4 == SYSCOOKIE_POOL.get(CookieIdx::FtruncateArg4).into()),
+                        scmp_cmp!($arg5 == SYSCOOKIE_POOL.get(CookieIdx::FtruncateArg5).into()),
                     ]);
                 }
 
@@ -296,12 +313,12 @@ impl AesWorker {
             Ok(syscall) => {
                 // Secure using syscall argument cookies.
                 let mut rules = vec![];
-                if !flags.allow_unsafe_nocookie() {
+                if restrict_cookie {
                     rules.extend(&[
                         #[cfg(target_arch = "x86")]
-                        scmp_cmp!($arg3 == (*FTRUNCATE64_COOKIE_ARG3).into()),
-                        scmp_cmp!($arg4 == (*FTRUNCATE64_COOKIE_ARG4).into()),
-                        scmp_cmp!($arg5 == (*FTRUNCATE64_COOKIE_ARG5).into()),
+                        scmp_cmp!($arg3 == SYSCOOKIE_POOL.get(Ftruncate64Arg3).into()),
+                        scmp_cmp!($arg4 == SYSCOOKIE_POOL.get(CookieIdx::Ftruncate64Arg4).into()),
+                        scmp_cmp!($arg5 == SYSCOOKIE_POOL.get(CookieIdx::Ftruncate64Arg5).into()),
                     ]);
                 }
 
@@ -326,10 +343,10 @@ impl AesWorker {
         if let Some(syscall) = SYS_ACCEPT4.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
             // Secure using syscall argument cookies.
             let mut rules = vec![];
-            if !flags.allow_unsafe_nocookie() {
+            if restrict_cookie {
                 rules.extend(&[
-                    scmp_cmp!($arg4 == (*ACCEPT4_COOKIE_ARG4).into()),
-                    scmp_cmp!($arg5 == (*ACCEPT4_COOKIE_ARG5).into()),
+                    scmp_cmp!($arg4 == SYSCOOKIE_POOL.get(CookieIdx::Accept4Arg4).into()),
+                    scmp_cmp!($arg5 == SYSCOOKIE_POOL.get(CookieIdx::Accept4Arg5).into()),
                 ]);
             }
 
@@ -351,99 +368,28 @@ impl AesWorker {
             }
         }
 
-        // Restrict pipe2(2) flags.
-        // Restrict pipe2(2) using syscall argument cookies.
-        let sysname = "pipe2";
-        #[expect(clippy::cast_sign_loss)]
-        #[expect(clippy::useless_conversion)]
-        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
-            let mut rules = vec![scmp_cmp!($arg1 == OFlag::O_CLOEXEC.bits() as u64)];
-            if !flags.allow_unsafe_nocookie() {
-                rules.extend(&[
-                    scmp_cmp!($arg2 == (*PIPE2_COOKIE_ARG2).into()),
-                    scmp_cmp!($arg3 == (*PIPE2_COOKIE_ARG3).into()),
-                    scmp_cmp!($arg4 == (*PIPE2_COOKIE_ARG4).into()),
-                    scmp_cmp!($arg5 == (*PIPE2_COOKIE_ARG5).into()),
-                ]);
-            }
-
-            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
-        } else {
-            info!("ctx": "confine", "op": "allow_aes_syscall",
-                "msg": format!("invalid or unsupported syscall {sysname}"));
-        }
+        // Restrict pipe2(2) flags, and use syscall argument cookies.
+        confine_scmp_pipe2(&mut ctx, restrict_cookie, OFlag::O_CLOEXEC)?;
 
         // Allow safe fcntl(2) utility calls.
-        for sysname in ["fcntl", "fcntl64"] {
-            let syscall = match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => syscall,
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_aes_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                    continue;
-                }
-            };
-
-            for op in AES_FCNTL_OPS {
-                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == *op)])?;
-            }
-        }
+        confine_scmp_fcntl(&mut ctx, AES_FCNTL_OPS)?;
 
         // Allow safe prctl(2) operations.
-        let sysname = "prctl";
-        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
-            for (_, op) in AES_PRCTL_OPS {
-                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == *op)])?;
-            }
-        } else {
-            info!("ctx": "confine", "op": "allow_aes_syscall",
-                "msg": format!("invalid or unsupported syscall {sysname}"));
-        }
+        confine_scmp_prctl(&mut ctx, AES_PRCTL_OPS)?;
 
         // Deny installing new signal handlers for {rt_,}sigaction(2).
-        for sysname in ["sigaction", "rt_sigaction"] {
-            let syscall = match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => syscall,
-                Err(_) => {
-                    crate::info!("ctx": "confine", "op": "allow_main_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                    continue;
-                }
-            };
-
-            // Installs a signal handler if first argument is non-NULL.
-            // We deny this case, but allow returning the current handler.
-            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == 0)])?;
-        }
+        confine_scmp_sigaction(&mut ctx)?;
 
         // Allow safe system calls.
-        for sysname in AES_SYSCALLS.iter().chain(VDSO_SYSCALLS) {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_aes_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
-        // Allow futex system calls.
-        for sysname in FUTEX_SYSCALLS {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_aes_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
-        // Allow getid system calls.
-        for sysname in GET_ID_SYSCALLS {
+        //
+        // KCOV_SYSCALLS is empty in case `kcov` feature is disabled.
+        for sysname in AES_SYSCALLS
+            .iter()
+            .chain(FUTEX_SYSCALLS)
+            .chain(GETID_SYSCALLS)
+            .chain(KCOV_SYSCALLS)
+            .chain(VDSO_SYSCALLS)
+        {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
@@ -456,8 +402,8 @@ impl AesWorker {
         }
 
         // Allow UID/GID changing system calls as necessary.
-        let safe_setuid = flags.allow_safe_setuid();
-        let safe_setgid = flags.allow_safe_setgid();
+        let safe_setuid = options.allow_safe_setuid();
+        let safe_setgid = options.allow_safe_setgid();
         if safe_setuid || safe_setgid {
             scmp_add_setid_rules(
                 "aes",
@@ -478,21 +424,37 @@ impl AesWorker {
             .name("syd_aes".to_string())
             .stack_size(AES_STACK_SIZE)
             .spawn(move || {
+                // SAFETY:
+                // 1. We use exit_group(2) here to bail, because this unsharing
+                //    is a critical safety feature.
+                // 2. CLONE_FILES cannot be unshared because this thread must share
+                //    file descriptors with the emulator threads.
+                if let Err(errno) = unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_SYSVSEM) {
+                    alert!("ctx": "boot", "op": "unshare_aes_thread",
+                        "msg": format!("failed to unshare(CLONE_FS|CLONE_SYSVSEM): {errno}"),
+                        "err": errno as i32);
+                    std::process::exit(101);
+                }
+
                 // Honour dry-run when exporting.
                 let dry_run =
-                    env::var_os(ENV_SKIP_SCMP).is_some() || ExportMode::from_env().is_some();
+                    secure_getenv(ENV_SKIP_SCMP).is_some() || ExportMode::from_env().is_some();
 
                 if !dry_run {
                     // Load the filter immediately.
                     // Logging is permitted as long as the `pid` key is unused.
                     // See prepare_confine for more information.
-                    let ctx =
-                        Self::prepare_confine(self.flags, &self.transit_uids, &self.transit_gids)?;
+                    let ctx = Self::prepare_confine(
+                        self.options,
+                        &self.transit_uids,
+                        &self.transit_gids,
+                        false,
+                    )?;
                     ctx.load()?;
 
-                    let safe_setid = self
-                        .flags
-                        .intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
+                    let safe_setid = self.options.intersects(
+                        Options::OPT_ALLOW_SAFE_SETUID | Options::OPT_ALLOW_SAFE_SETGID,
+                    );
                     info!("ctx": "confine", "op": "confine_aes_thread",
                         "msg": format!("AES thread confined with{} SROP mitigation",
                             if safe_setid { "out" } else { "" }));
