@@ -9,12 +9,15 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
-use std::sync::{
-    atomic::{AtomicBool, Ordering},
-    Arc, RwLock,
+use std::{
+    os::fd::AsFd,
+    sync::{
+        atomic::{AtomicBool, Ordering},
+        Arc, RwLock,
+    },
 };
 
-use libseccomp::{ScmpAction, ScmpFilterContext, ScmpSyscall};
+use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
     sys::{
@@ -28,8 +31,9 @@ use crate::{
     compat::{waitid, WaitStatus},
     config::*,
     confine::{
-        confine_scmp_execveat, confine_scmp_faccessat2, confine_scmp_fcntl, confine_scmp_ioctl_syd,
+        confine_scmp_execveat, confine_scmp_faccessat2, confine_scmp_ioctl_syd,
         confine_scmp_madvise, confine_scmp_open, confine_scmp_openat, confine_scmp_openat2,
+        confine_scmp_pidfd_getfd, confine_scmp_pidfd_open, confine_scmp_pidfd_send_signal,
         confine_scmp_prctl, confine_scmp_sigaction, confine_scmp_write, scmp_add_setid_rules,
     },
     err::SydResult,
@@ -38,7 +42,7 @@ use crate::{
         exec::sysevent_exec, exit::sysevent_exit, fork::sysevent_fork, scmp::sysevent_scmp,
         sig::sysevent_sig, sysx::sysevent_sysx,
     },
-    sandbox::{Flags, Sandbox},
+    sandbox::{LockState, Options, Sandbox, SandboxGuard},
     workers::WorkerCache,
 };
 
@@ -63,14 +67,20 @@ impl Tracer {
     }
 
     /// Run the ptrace(2) loop. This is the main entry point.
-    pub(crate) fn run(self, child: Pid, wait_all: bool) -> SydResult<u8> {
+    pub(crate) fn run<Fd: AsFd>(
+        self,
+        child_pfd: Fd,
+        child_pid: Pid,
+        wait_all: bool,
+    ) -> SydResult<u8> {
         // Wait in a loop and push WaitStatus into queue.
         let mut xcode = 127;
         loop {
             match waitid(Id::All, WaitPidFlag::WEXITED | WaitPidFlag::__WNOTHREAD) {
                 Ok(WaitStatus::Exited(pid, exit_code)) => {
-                    self.cache.del_pid(pid);
-                    if pid == child {
+                    let is_child = pid == child_pid;
+                    self.handle_exit(pid, is_child, wait_all);
+                    if is_child {
                         xcode = exit_code;
                         if !wait_all {
                             break;
@@ -79,7 +89,7 @@ impl Tracer {
                 }
                 Ok(WaitStatus::Signaled(pid, signal, _core)) => {
                     self.cache.del_pid(pid);
-                    if pid == child {
+                    if pid == child_pid {
                         xcode = 128_i32.saturating_add(signal);
                         if !wait_all {
                             break;
@@ -98,6 +108,33 @@ impl Tracer {
             }
         }
 
+        // Ghost mode:
+        // 1. If should_exit was set by one of the emulator threads
+        //    before main thread could collect child's exit status,
+        //    do a final blocking wait to get the correct exit code.
+        // 2. Use the pid file descriptor to avoid PID recycling.
+        if xcode == 127 {
+            let flags = WaitPidFlag::WEXITED | WaitPidFlag::__WNOTHREAD;
+            loop {
+                match waitid(Id::PIDFd(child_pfd.as_fd()), flags) {
+                    Ok(WaitStatus::Exited(_, exit_code)) => {
+                        xcode = exit_code;
+                        break;
+                    }
+                    Ok(WaitStatus::Signaled(_, signal, _)) => {
+                        xcode = 128_i32.saturating_add(signal);
+                        break;
+                    }
+                    Ok(status) => {
+                        unreachable!("BUG: final waitid returned {status:?}, report a bug!")
+                    }
+                    Err(Errno::EINTR | Errno::EAGAIN) => {}
+                    Err(Errno::ECHILD) => break,
+                    Err(errno) => return Err(errno.into()),
+                }
+            }
+        }
+
         // Set should_exit to true and notify the syd_aes thread.
         self.should_exit.store(true, Ordering::Relaxed);
         if let Some(ref crypt_map) = self.cache.crypt_map {
@@ -148,7 +185,7 @@ impl Tracer {
                 sysevent_fork(pid, &self.sandbox);
             }
             WaitStatus::PtraceEvent(pid, libc::SIGTRAP, libc::PTRACE_EVENT_EXEC) => {
-                sysevent_exec(pid, &self.cache, &self.sandbox);
+                sysevent_exec(pid, &self.sandbox);
             }
             WaitStatus::PtraceEvent(pid, libc::SIGTRAP, libc::PTRACE_EVENT_EXIT) => {
                 sysevent_exit(pid, &self.cache, &self.sandbox);
@@ -157,16 +194,42 @@ impl Tracer {
         }
     }
 
+    fn handle_exit(&self, pid: Pid, is_child: bool, wait_all: bool) {
+        // Remove cache entries which belong to this PID.
+        self.cache.del_pid(pid);
+
+        // We're done if:
+        // (a) This is not the eldest process.
+        // (b) trace/exit_wait_all is not set and we'll exit shortly.
+        if !is_child || !wait_all {
+            return;
+        }
+
+        // Lock sandbox immediately if we're in lock:exec.
+        // The eldest child exited and sandbox can no longer
+        // be edited. Let's assert that.
+        let mut sandbox =
+            SandboxGuard::Write(self.sandbox.write().unwrap_or_else(|err| err.into_inner()));
+        if sandbox.lock == Some(LockState::Exec) {
+            // Panic is the only option here on errors.
+            #[expect(clippy::disallowed_methods)]
+            sandbox
+                .lock(LockState::Set)
+                .expect("BUG: failed to lock sandbox, report a bug!");
+        }
+    }
+
     /// Prepare to confine the Tracer threads.
+    #[expect(clippy::cognitive_complexity)]
     pub(crate) fn prepare_confine(
-        flags: Flags,
+        options: Options,
         transit_uids: &[(Uid, Uid)],
         transit_gids: &[(Gid, Gid)],
     ) -> SydResult<ScmpFilterContext> {
-        let ssb = flags.allow_unsafe_exec_speculative();
-        let restrict_cookie = !flags.allow_unsafe_nocookie();
-        let safe_setuid = flags.allow_safe_setuid();
-        let safe_setgid = flags.allow_safe_setgid();
+        let ssb = options.allow_unsafe_exec_speculative();
+        let restrict_cookie = !options.allow_unsafe_nocookie();
+        let safe_setuid = options.allow_safe_setuid();
+        let safe_setgid = options.allow_safe_setgid();
         let safe_setid = safe_setuid || safe_setgid;
 
         let mut ctx = ScmpFilterContext::new(ScmpAction::KillProcess)?;
@@ -210,7 +273,26 @@ impl Tracer {
         confine_scmp_madvise(&mut ctx)?;
 
         // Allow safe fcntl(2) utility calls.
-        confine_scmp_fcntl(&mut ctx, MAIN_FCNTL_OPS)?;
+        for sysname in ["fcntl", "fcntl64"] {
+            let syscall = match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) => syscall,
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_gdb_syscall",
+                        "msg": format!("invalid or unsupported syscall {sysname}"));
+                    continue;
+                }
+            };
+
+            // TODO: Figure out what fcntl(2) ops are needed for KCOV.
+            if cfg!(feature = "kcov") {
+                ctx.add_rule(ScmpAction::Allow, syscall)?;
+                continue;
+            }
+
+            for op in MAIN_FCNTL_OPS {
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == *op)])?;
+            }
+        }
 
         // Allow safe prctl(2) operations.
         confine_scmp_prctl(&mut ctx, MAIN_PRCTL_OPS)?;
@@ -224,11 +306,13 @@ impl Tracer {
 
         // Allow safe system calls.
         //
-        // Note, `PROF_SYSCALLS` is empty in case `prof` feature is disabled.
+        // KCOV_SYSCALLS is empty in case `kcov` feature is disabled.
+        // PROF_SYSCALLS is empty in case `prof` feature is disabled.
         for sysname in MAIN_SYSCALLS
             .iter()
             .chain(FUTEX_SYSCALLS)
             .chain(GETID_SYSCALLS)
+            .chain(KCOV_SYSCALLS)
             .chain(PROF_SYSCALLS)
             .chain(VDSO_SYSCALLS)
         {
@@ -246,6 +330,11 @@ impl Tracer {
         // Allow faccessat2(2) system call.
         confine_scmp_faccessat2(&mut ctx, restrict_cookie)?;
 
+        // pidfd family system calls may be used only with syscall argument cookies.
+        confine_scmp_pidfd_getfd(&mut ctx, restrict_cookie)?;
+        confine_scmp_pidfd_open(&mut ctx, restrict_cookie)?;
+        confine_scmp_pidfd_send_signal(&mut ctx, restrict_cookie)?;
+
         // Allow UID/GID changing system calls as necessary.
         if safe_setid {
             scmp_add_setid_rules(
