@@ -43,10 +43,12 @@ use crate::{
     alert,
     compat::{epoll_ctl_mod_safe, epoll_ctl_safe, MsgFlags},
     config::*,
-    confine::{confine_scmp_madvise, confine_scmp_write, scmp_add_setid_rules},
+    confine::{
+        confine_scmp_fcntl, confine_scmp_madvise, confine_scmp_prctl, confine_scmp_write,
+        confine_scmp_wx_syd, scmp_add_setid_rules,
+    },
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
-    fd::closeexcept,
-    fs::peer_cred,
+    fd::{closeexcept, peer_creds},
     hash::SydHashMap,
     info,
     landlock::Errata,
@@ -54,7 +56,7 @@ use crate::{
     path::{XPath, XPathBuf},
     retry::retry_on_eintr,
     rng::duprand,
-    sandbox::{Flags, Sandbox, LINE_MAX},
+    sandbox::{Options, Sandbox, LINE_MAX},
 };
 
 const IPC_ACK: &[u8] = b"{\"err\":0,\"msg\":\"ACK\"}\n";
@@ -76,8 +78,8 @@ pub(crate) struct IpcWorker {
     pub(crate) sock: Option<UnixListener>,
     // Reference to the Sandbox to be configured
     sandbox: Option<Arc<RwLock<Sandbox>>>,
-    // Sandbox flags specified at startup.
-    flags: Flags,
+    // Sandbox options specified at startup.
+    options: Options,
     // Credentials for authentication.
     creds: (Option<Uid>, Option<Gid>),
 
@@ -234,12 +236,12 @@ impl IpcWorker {
         addr: &XPath,
         uid: Option<Uid>,
         gid: Option<Gid>,
-        flags: Flags,
+        options: Options,
         transit_uids: &[(Uid, Uid)],
         transit_gids: &[(Gid, Gid)],
     ) -> Self {
         IpcWorker {
-            flags,
+            options,
             creds: (uid, gid),
             addr: Some(addr.to_owned()),
             epoll: None,
@@ -328,7 +330,7 @@ impl IpcWorker {
     #[expect(clippy::cognitive_complexity)]
     pub(crate) fn prepare_confine(
         epoll_fd: RawFd,
-        flags: Flags,
+        options: Options,
         transit_uids: &[(Uid, Uid)],
         transit_gids: &[(Gid, Gid)],
         dry_run: bool,
@@ -358,7 +360,7 @@ impl IpcWorker {
 
         // Disable Speculative Store Bypass mitigations
         // with trace/allow_unsafe_exec_speculative:1
-        ctx.set_ctl_ssb(flags.allow_unsafe_exec_speculative())?;
+        ctx.set_ctl_ssb(options.allow_unsafe_exec_speculative())?;
 
         // We kill for bad system call and bad arch.
         ctx.set_act_badarch(ScmpAction::KillProcess)?;
@@ -421,49 +423,13 @@ impl IpcWorker {
         }
 
         // Allow safe fcntl(2) utility calls.
-        for sysname in ["fcntl", "fcntl64"] {
-            let syscall = match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => syscall,
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_ipc_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                    continue;
-                }
-            };
-
-            for op in IPC_FCNTL_OPS {
-                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == *op)])?;
-            }
-        }
+        confine_scmp_fcntl(&mut ctx, IPC_FCNTL_OPS)?;
 
         // Allow safe prctl(2) operations.
-        let sysname = "prctl";
-        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
-            for (_, op) in IPC_PRCTL_OPS {
-                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == *op)])?;
-            }
-        } else {
-            info!("ctx": "confine", "op": "allow_ipc_syscall",
-                "msg": format!("invalid or unsupported syscall {sysname}"));
-        }
+        confine_scmp_prctl(&mut ctx, IPC_PRCTL_OPS)?;
 
         // Prevent executable memory.
-        const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
-        for sysname in ["mmap", "mmap2", "mprotect"] {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule_conditional(
-                        ScmpAction::Allow,
-                        syscall,
-                        &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
-                    )?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_ipc_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
+        confine_scmp_wx_syd(&mut ctx)?;
 
         // Allow writes to the log-fd.
         // No proc_pid_mem(5) access required here.
@@ -473,33 +439,15 @@ impl IpcWorker {
         confine_scmp_madvise(&mut ctx)?;
 
         // Allow safe system calls.
-        for sysname in IPC_SYSCALLS.iter().chain(VDSO_SYSCALLS) {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_ipc_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
-        // Allow futex system calls.
-        for sysname in FUTEX_SYSCALLS {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_ipc_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
-        // Allow getid system calls.
-        for sysname in GET_ID_SYSCALLS {
+        //
+        // KCOV_SYSCALLS is empty in case `kcov` feature is disabled.
+        for sysname in IPC_SYSCALLS
+            .iter()
+            .chain(FUTEX_SYSCALLS)
+            .chain(GETID_SYSCALLS)
+            .chain(KCOV_SYSCALLS)
+            .chain(VDSO_SYSCALLS)
+        {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
@@ -512,8 +460,8 @@ impl IpcWorker {
         }
 
         // Allow UID/GID changing system calls as necessary.
-        let safe_setuid = flags.allow_safe_setuid();
-        let safe_setgid = flags.allow_safe_setgid();
+        let safe_setuid = options.allow_safe_setuid();
+        let safe_setgid = options.allow_safe_setgid();
         if safe_setuid || safe_setgid {
             scmp_add_setid_rules(
                 "ipc",
@@ -603,7 +551,7 @@ impl IpcWorker {
         // because this confinement is a critical safety feature.
         let ctx = match Self::prepare_confine(
             epoll.0.as_raw_fd(),
-            self.flags,
+            self.options,
             &self.transit_uids,
             &self.transit_gids,
             false,
@@ -702,7 +650,7 @@ impl IpcWorker {
                     epoll_ctl_safe(&epoll.0, fd, Some(event))?;
 
                     // Create Connection state.
-                    let creds = peer_cred(&stream)?;
+                    let creds = peer_creds(&stream)?;
                     let mut conn = Connection::new(creds, stream);
 
                     // In prompt mode, send initial prompt.
@@ -1038,8 +986,8 @@ impl IpcWorker {
 
         // Close connection immediately if sandbox is locked.
         // This also results in thread-exit.
-        if sandbox.locked() {
-            return Err(Errno::EBUSY);
+        if sandbox.is_locked() {
+            return Err(Errno::EPERM);
         }
 
         // Intercept display commands and send to socket
@@ -1060,8 +1008,8 @@ impl IpcWorker {
 
         // Close connection immediately if sandbox got locked.
         // This also results in thread-exit.
-        if sandbox.locked() {
-            return Err(Errno::EBUSY);
+        if sandbox.is_locked() {
+            return Err(Errno::EPERM);
         }
 
         match result {
